import argparse, pathlib, sys, json, datetime

from .simpledevice import SimpleCtrDevice
from .soap.manager import CtrSoapManager
from .soap import ias, ecs, helpers
from .ninja import NinjaManager, NinjaException
from .act import ActSimpleObj
from .title import Title, MediaType
from ..nintendowifi.soapenvelopebase import SoapCodeError
from .secureinfo import SecureInfo
from .cdn import CDN
from .cdn_helpers import CiaCDNBuilder
from . import regionaldata, keys, constants

def _print_unhandled_exceptions(func):
	def inner(*args, **kwds):
		try:
			return func(*args, **kwds)
		except Exception as e:
			print(f'An exception occurred:\n - {type(e).__name__}:\n  - {e}')
			raise # in the future, use this for logging properly, for now
	return inner

def _print_soapcode_exceptions(func):
	def inner(*args, **kwds):
		try:
			return func(*args, **kwds)
		except SoapCodeError as e:
			print('A SoapCodeError was raised:')
			print(f' Exception args: {str(e)}')
			print(f' Soap error code: {e.soaperrorcode}')
			print(f' Soap error message: {e.soaperrormessage}')
	return inner

@_print_unhandled_exceptions
def _setup_constants(parser, parsed_args):
	if parsed_args.aes_constant_c:
		c = None
		try:
			c = bytes.fromhex(parsed_args.aes_constant_c)
		except Exception:
			pass
		constants._load_const_c(c)
	if parsed_args.ssl:
		constants._configure_certs(parsed_args.ssl.read(4*1024*1024))
	if parsed_args.cfg:
		constants._load_cfg_rsa_keys(parsed_args.cfg.read(4*1024*1024))
	if parsed_args.act:
		constants._load_act_secrets(parsed_args.act.read(4*1024*1024))
	if parsed_args.process9:
		constants._load_p9_keys(parsed_args.process9.read(4*1024*1024))
	if parsed_args.enc_clcert_cert and parsed_args.enc_clcert_key:
		constants._configure_cli_cert_and_key(
			parsed_args.enc_clcert_cert.read(64*1024),
			parsed_args.enc_clcert_key.read(64*1024)
		)

	print("If any valid data was found, respective constants were setup.")

@_print_unhandled_exceptions
def _gen_json(parser, parsed_args):
	print("Generating json...")

	SimpleCtrDevice.generate_new_json(
		otp_fp=parsed_args.otp,
		secureinfo_fp=parsed_args.secureinfo,
		serialnumber=parsed_args.serialnumber,
		msed_fp=parsed_args.msed,
		model_override=parsed_args.model,
		region=parsed_args.region,
		country=parsed_args.country,
		language=parsed_args.language,
		json_file=parsed_args.out
	)

	print("Complete!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _move_account(parser, parsed_args):
	print("Initializing source console...")
	source = SimpleCtrDevice(json_file=parsed_args.source)
	soap_source = CtrSoapManager(source, False)

	print("Initializing target console...")
	target = SimpleCtrDevice(json_file=parsed_args.target)
	soap_target = CtrSoapManager(target, False)

	helpers.CtrSoapUseSystemApps(soap_source, helpers.SysApps.SYSTRANSFER)
	helpers.CtrSoapUseSystemApps(soap_target, helpers.SysApps.SYSTRANSFER)

	print("Initializing console sessions...")
	helpers.CtrSoapSessionConnect(soap_source)
	helpers.CtrSoapSessionConnect(soap_target)

	print("Saving updated sessions...")
	source.serialize_json(json_file=parsed_args.source)
	target.serialize_json(json_file=parsed_args.target)

	print("Checking if we can move account...")
	movestatus = ias.MoveAccount(
		soap_source,
		soap_target.device_id,
		soap_target.account_id,
		soap_target.st_token,
		True
	)

	print("Performing move...")
	movestatus = ias.MoveAccount(
		soap_source,
		soap_target.device_id,
		soap_target.account_id,
		soap_target.st_token,
		False
	)

	print("Complete!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _move_nnid(parser, parsed_args):
	print("Initializing source console...")
	source = SimpleCtrDevice(json_file=parsed_args.source)
	soap_source = CtrSoapManager(source, False)

	print("Initializing target console...")
	target = SimpleCtrDevice(json_file=parsed_args.target)
	soap_target = CtrSoapManager(target, False)

	helpers.CtrSoapUseSystemApps(soap_source, helpers.SysApps.SYSTRANSFER)
	helpers.CtrSoapUseSystemApps(soap_target, helpers.SysApps.SYSTRANSFER)

	print("Initializing console sessions...")
	helpers.CtrSoapSessionConnect(soap_source)
	helpers.CtrSoapSessionConnect(soap_target)

	print("Saving updated sessions...")
	source.serialize_json(json_file=parsed_args.source)
	target.serialize_json(json_file=parsed_args.target)

	print("Attempt to move NNID")
	act_source = ActSimpleObj(
		source,
		source.titles(MediaType.NAND)[
			source.titles(MediaType.NAND).index(soap_source.application_id)
		]
	)
	act_target = ActSimpleObj(
		target,
		target.titles(MediaType.NAND)[
			target.titles(MediaType.NAND).index(soap_target.application_id)
		]
	)

	http, response = act_source.devices_current_migrations_commit()
	if (http // 100) == 2 and not response.had_errors:
		http, response = act_source.devices_current_migrations(delete=True)
	if (http // 100) == 2 and not response.had_errors:
		http, response = act_source.devices_current_migrations(
			delete=False,
			serialnumber=target.serial_no,
			deviceid=target.device_id
		)
	if (http // 100) == 2 and not response.had_errors:
		http, response = act_target.devices_current_migrations_commit()
	if (http // 100) != 2 or response.had_errors:
		print("Fail!")
		print(f"http response: {http}")
		print("")
		for i in response.errors:
			print("account api error response:")
			print(f"- cause: {i.cause}")
			print(f"- code: {i.code}")
			print(f"- message: {i.message}")
			print("")
		return

	print("Complete!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _move_check_account(parser, parsed_args):
	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Checking registry...")
	helpers.CtrSoapCheckRegister(soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	acct_attributes = ias.GetAccountAttributesByProfile(soap_device, 'MOVE_ACCT')

	moved_times = None
	moved_last_time = None

	for i in acct_attributes.accountattributes:
		if   i[0] == 'MoveAccountTimes':
			moved_times = int(i[1]) if i[1] else 0
		elif i[0] == 'MoveAccountLastMovedDate':
			moved_last_time = int(i[1]) if i[1] else 0

	utc_0 = datetime.datetime.utcfromtimestamp(0)

	server_time = utc_0 + datetime.timedelta(milliseconds=acct_attributes.timestamp)
	if moved_last_time is not None:
		moved_last_time = utc_0 + datetime.timedelta(milliseconds=moved_last_time)
		time_ready_for_new_move = moved_last_time + datetime.timedelta(days=7)

	if moved_times is not None:
		print(f"Moved times: {moved_times}")
	else:
		print("No moved times info!")

	if moved_last_time is not None:
		print(f"Moved last date: {moved_last_time.strftime('%a, %d %b %Y %H:%M:%S UTC')}")
	else:
		print("No moved last date info!")

	print(f"Server time at response: {server_time.strftime('%a, %d %b %Y %H:%M:%S UTC')}")

	if moved_last_time is not None and time_ready_for_new_move > server_time:
		delta = time_ready_for_new_move - server_time
		d = delta.days
		h = delta.seconds // 3600
		m = delta.seconds // 60 % 60
		s = delta.seconds % 60
		μs = delta.microseconds
		print(f"Ready for transfer in {d}d {h}h {m}min {s}s {μs}μs")
		print(f"Ready at: {time_ready_for_new_move.strftime('%a, %d %b %Y %H:%M:%S UTC')}")
	else:
		print("Ready for transfer!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _get_ivs(parser, parsed_args):
	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Initializing console session with forced get ivs...")
	helpers.CtrSoapSessionConnect(soap_device, forced_ivs_remote_import=True)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	print("Complete!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _set_ivs(parser, parsed_args):
	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Initializing console session with forced set ivs...")
	helpers.CtrSoapSessionConnect(soap_device, forced_ivs_send=True)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	print("Complete!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _recover_ivs_otp(parsed, parsed_args):
	print("Generating json serialless in memory...")

	c = SimpleCtrDevice.generate_new_json(
		otp_fp=parsed_args.otp,
		serialnumber='XXX12345678',
		region=parsed_args.region,
		country=parsed_args.country,
		language=parsed_args.language
	)

	print("Initializing console...")
	device = SimpleCtrDevice(json_string=c)
	soap_device = CtrSoapManager(device, False)

	print("Attempt movable.sed recovery...")
	reginfo = ias.GetRegistrationInfo(soap_device, ias.GetChallenge(soap_device).challenge)

	if not reginfo.is_registered:
		print("Cannot recover, no registry.")
		return

	elif reginfo.devicetokenexpired:
		print("Cannot recover, device token expired.")
		return

	soap_device.register_account('R', reginfo.devicetoken, reginfo.accountid, reginfo.country, reginfo.extaccountid, reginfo.devicecode)

	remote_ivs = ias.GetIVSData(soap_device).ivsdata

	if not remote_ivs:
		print("Cannot recover, no IVS found.")
		return

	soap_device.import_ivs(remote_ivs)

	print("Saving Movable.sed...")
	with open(parsed_args.out, 'wb') as o:
		o.write(bytes(device.msed))

	print("Complete!")

def _run_unregister(device, soap_device):
	try:
		ias.Unregister(soap_device, ias.GetChallenge(soap_device).challenge)
		soap_device.unregister_account()
		virtual = False
	except SoapCodeError as e:
		if e.soaperrorcode != 434:
			raise
		virtual = True

	if virtual:
		print("Virtual account link! Attempt detach by error...")
		device.reboot()

		print("Initializing console session...")
		helpers.CtrSoapUseSystemApps(soap_device, helpers.SysApps.SYSTRANSFER)
		helpers.CtrSoapSessionConnect(soap_device)

		device_ninja = NinjaManager(soap_device)
		try:
			device_ninja.open_without_nna()
		except NinjaException as e:
			if e.errorcode != 3136:
				raise

		device.reboot()

		print("Initializing console...")
		helpers.CtrSoapUseSystemApps(soap_device, helpers.SysApps.ESHOP)

		print("Checking registry...")
		helpers.CtrSoapCheckRegister(soap_device)

		if soap_device.account_status != 'U':
			print("Unregister...")
			ias.Unregister(soap_device, ias.GetChallenge(soap_device).challenge)
			soap_device.unregister_account()
		else:
			print("Unregistered!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _del_eshop(parser, parsed_args):
	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Checking registry...")
	helpers.CtrSoapCheckRegister(soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	if soap_device.account_status == 'U':
		print("Console already does not have EShop account.")
		return

	device.reboot()

	print("Initializing console session...")
	helpers.CtrSoapUseSystemApps(soap_device, helpers.SysApps.ESHOP)
	helpers.CtrSoapSessionConnect(soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	print("Unregister...")
	_run_unregister(device, soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	print("Complete!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _eshop_region_change(parser, parsed_args):
	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Checking registry...")
	helpers.CtrSoapCheckRegister(soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	if parsed_args.region == soap_device.region and soap_device.account_status != 'U':
		print("Console already in the desired region.")
		return

	device.reboot()

	if soap_device.account_status != 'U':
		print("Initializing console session...")
		helpers.CtrSoapUseSystemApps(soap_device, helpers.SysApps.ESHOP)
		helpers.CtrSoapSessionConnect(soap_device)

		print("Saving updated session...")
		device.serialize_json(json_file=parsed_args.console)

		print("Unregister...")
		_run_unregister(device, soap_device)

		print("Saving updated session...")
		device.serialize_json(json_file=parsed_args.console)

		device.reboot()

	soap_device.region_change(parsed_args.region, parsed_args.country, parsed_args.language)

	print("Initializing console session...")
	helpers.CtrSoapUseSystemApps(soap_device, helpers.SysApps.ESHOP)
	try:
		helpers.CtrSoapSessionConnect(soap_device)
	except SoapCodeError as e:
		if e.soaperrorcode == 602:
			print("We got soap error 602.")
			print("Region could not be changed.")
			print("Any existing eshop account was deleted in the process.")
			print("This console has titles attached to it on a different region.")
			print("System transfer to another console is needed to remove them.")
			print("System transfer without NNID transfer is enough.")
			print("NNID-only transfers do not work to fix.")
			return
		raise

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	print("Complete!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _list_etickets(parser, parsed_args):
	_format = parsed_args.format.lower()
	if _format not in ['text', 'json']:
		parser.error("Invalid output format.")
		return

	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Initializing console session...")
	helpers.CtrSoapUseSystemApps(soap_device, helpers.SysApps.ESHOP)
	helpers.CtrSoapSessionConnect(soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	accountetickets = ecs.AccountListETicketIds(soap_device)

	tivs = []
	for i in accountetickets.tivs:
		tivs.append(i[0])

	accountetickets = ecs.AccountGetETicketDetails(soap_device, tivs)

	out = parsed_args.out
	out = out.open('w', encoding='utf-8') if out else sys.stdout

	if _format == 'text':
		out.write(f"Total ETickets: {len(accountetickets.eticketinfos)}\n")
		for i in accountetickets.eticketinfos:
			out.write("ETicketInfo:\n")
			out.write(f" - TicketId: {i.ticketid}\n")
			out.write(f" - TitleId: {i.titleid:016X}\n")
			out.write(f" - Version: {i.version}\n")
			out.write(f" - FormatVersion: {i.formatversion}\n")
			out.write(f" - MigrationCount: {i.migratecount}\n")
			out.write(f" - MigrationLimit: {i.migratelimit}\n")
			out.write(f" - EstimatedSize: {i.estimatedsize}\n")
	elif _format == 'json':
		json.dump(
			{'ETicketInfos': accountetickets.eticketinfos},
			out,
			indent=2,
			default=lambda x: x.__dict__
		)
		print("")

	if out is not sys.stdout:
		out.close()

	print("Complete!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _check_register(parser, parsed_args):
	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Checking registry...")
	helpers.CtrSoapCheckRegister(soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	print(f"Account status: {soap_device.account_status}")
	if soap_device.account_status != 'U':
		print(f"Account register: {'Expired' if soap_device.register_expired else 'Valid'}")
	print(f"Current effective region: {soap_device.region}")
	print(f"Current effective country: {soap_device.country}")
	print(f"Current effective language: {soap_device.language}")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _download_etickets(parser, parsed_args):
	if parsed_args.ticket_id and not parsed_args.all:
		_tikids = sum(parsed_args.ticket_id, [])
		tikids = []
		for i in _tikids:
			try:
				tikids.append(int(i))
			except Exception:
				pass
		del _tikids
		if not tikids:
			print("No valid ticket ids received and --all is unset.")
			return

	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Initializing console session...")
	helpers.CtrSoapUseSystemApps(soap_device, helpers.SysApps.ESHOP)
	helpers.CtrSoapSessionConnect(soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	accountetickets = ecs.AccountListETicketIds(soap_device)

	account_tivs = []
	for i in accountetickets.tivs:
		account_tivs.append(i[0])

	if parsed_args.all:
		tikids = account_tivs
	else:
		tikids = list(set(tikids).intersection(account_tivs))

	if not tikids:
		print("No ticket ids were found to download.")
		return

	etiks = []
	certs = []

	print("Fetching Tickets...")

	try:
		parsed_args.out.mkdir(parents=True, exist_ok=True)
	except Exception:
		print("Failed to ensure output path.")
		return

	for i in tikids:
		# yes, AccountGetETickets can actually take multiple ids.
		# but 3ds does not do that.
		etik = ecs.AccountGetETickets(soap_device, [i])
		etiks += list(etik.etickets)
		certs += list(etik.certs)
		certs = list(set(etik.certs))

	print(f"Saving {len(etiks)} tickets and {len(certs)} certs...")

	for i in etiks:
		out_name = f"{i.title_id:016X}.{i.ticket_id}.{i.device_id:08X}.{i.eshop_id:08X}.tik"
		out_path = parsed_args.out / out_name
		try:
			with out_path.open('wb') as o:
				o.write(bytes(i) if parsed_args.decrypt else i.installable_export)
		except Exception as e:
			print("Failed to write " + out_name)
			print(type(e), e)

	for i in certs:
		out_name = f"{i.full_issuer}.bin"
		out_path = parsed_args.out / out_name
		try:
			with out_path.open('wb') as o:
				o.write(bytes(i))
		except Exception as e:
			print("Failed to write " + out_name)
			print(type(e), e)

	print("Done!")

@_print_unhandled_exceptions
@_print_soapcode_exceptions
def _download_title_tikid(parser, parsed_args):
	if parsed_args.ticket_id and not parsed_args.all:
		_tikids = sum(parsed_args.ticket_id, [])
		tikids = []
		for i in _tikids:
			try:
				tikids.append(int(i))
			except Exception:
				pass
		del _tikids
		if not tikids:
			print("No valid ticket ids received and --all is unset.")
			return

	print("Initializing console...")
	device = SimpleCtrDevice(json_file=parsed_args.console)
	soap_device = CtrSoapManager(device, False)

	print("Initializing console session...")
	helpers.CtrSoapUseSystemApps(soap_device, helpers.SysApps.ESHOP)
	helpers.CtrSoapSessionConnect(soap_device)

	print("Saving updated session...")
	device.serialize_json(json_file=parsed_args.console)

	accountetickets = ecs.AccountListETicketIds(soap_device)

	account_tivs = []
	for i in accountetickets.tivs:
		account_tivs.append(i[0])

	if parsed_args.all:
		tikids = account_tivs
	else:
		tikids = list(set(tikids).intersection(account_tivs))

	if not tikids:
		print("No ticket ids were found to download.")
		return

	etiks = []
	certs = {}

	print("Fetching tickets...")

	try:
		parsed_args.out.mkdir(parents=True, exist_ok=True)
	except Exception:
		print("Failed to ensure output path.")
		return

	for i in tikids:
		# yes, AccountGetETickets can actually take multiple ids.
		# but 3ds does not do that.
		etik = ecs.AccountGetETickets(soap_device, [i])
		etiks += list(etik.etickets)
		_certs = list(set(etik.certs))
		for j in _certs:
			certs[j.full_issuer] = j

	print(f"Downloading {len(etiks)} titles...")

	for i in etiks:
		cdn = CDN(
			i.title_id, 
			device_id=i.device_id,
			account_id=i.eshop_id,
			ticket=i,
			content_prefix=soap_device.get_url_by_identifier('content_prefix'),
			uncached_content_prefix=soap_device.get_url_by_identifier('uncached_content_prefix'),
			system_content_prefix=soap_device.get_url_by_identifier('system_content_prefix'),
			system_uncached_content_prefix=soap_device.get_url_by_identifier('system_uncached_content_prefix'),
			is_dev=device.is_dev
		)

		cert = certs.get(i.issuer, None)
		while cert:
			cdn.add_additional_cert(cert)
			cert = certs.get(cert.issuer, None)

		out_name = f"{i.title_id:016X}.{i.ticket_id}.{i.device_id:08X}.{i.eshop_id:08X}.cia"
		out_path = parsed_args.out / out_name

		print(f"Downloading to {out_name}")
		try:
			helper = CiaCDNBuilder(str(out_path), parsed_args.keep_going)
			if not cdn.download(
				helper,
				right_check=not parsed_args.ignore_rights,
				content_error_continue=parsed_args.keep_going
			):
				print("Failed to download to " + out_name)
				print("This may happen with connection errors, missing content or other unexpected reasons")
		except Exception as e:
			print("Failed to download to " + out_name)
			print(type(e), e)
			raise

	print("Done!")

def _main(args = None):
	parser = argparse.ArgumentParser(prog='cleaninty.ctr')
	parser.add_argument(
		'--boot9',
		help='3DS\'S boot9 for keys if not on the environment',
		type=str
	)

	subparsers = parser.add_subparsers(required=True, metavar='command')

	constants_parser = subparsers.add_parser('SetupConstants', help='Setup various used constants to environment')
	constants_parser.add_argument(
		'--aes-constant-c',
		help='Hardware AES Constant C as an hex string',
		type=str
	)
	constants_parser.add_argument(
		'--ssl',
		help='Path to SSL Module\'s code.bin',
		type=argparse.FileType('rb')
	)
	constants_parser.add_argument(
		'--cfg',
		help='Path to CFG Module\'s code.bin',
		type=argparse.FileType('rb')
	)
	constants_parser.add_argument(
		'--act',
		help='Path to ACT Module\'s code.bin',
		type=argparse.FileType('rb')
	)
	constants_parser.add_argument(
		'--process9',
		help='Path to Process9\'s code.bin',
		type=argparse.FileType('rb')
	)
	constants_parser.add_argument(
		'--enc-clcert-cert',
		help='Path to ClCertA\'s ctr-common-1-cert.bin',
		type=argparse.FileType('rb')
	)
	constants_parser.add_argument(
		'--enc-clcert-key',
		help='Path to ClCertA\'s ctr-common-1-key.bin',
		type=argparse.FileType('rb')
	)
	constants_parser.set_defaults(func=_setup_constants)

	gen_parser = subparsers.add_parser('GenJson', help='Generate Console JSON Object')
	gen_parser.add_argument(
		'--otp',
		help='Console\'s encrypted or decrypted otp file',
		required=True,
		type=argparse.FileType('rb')
	)
	gen_secinfo_parser = gen_parser.add_mutually_exclusive_group(required=True)
	gen_secinfo_parser.add_argument(
		'--secureinfo',
		help='Console\'s SecureInfo_A/B file',
		type=argparse.FileType('rb')
	)
	gen_secinfo_parser.add_argument(
		'--serialnumber',
		help='Console\'s serial number if SecureInfo not given',
		type=SecureInfo.validate_serial
	)
	gen_parser.add_argument(
		'--msed',
		help='Console\'s movable.sed, only add if you need, this will interfere with IVS sync',
		type=argparse.FileType('rb')
	)
	gen_parser.add_argument(
		'--model', '-m',
		help='Console\'s model override',
		type=str
	)
	gen_parser.add_argument(
		'--region', '-r',
		help='Console\'s effective region',
		type=regionaldata.Region.get_region
	)
	gen_parser.add_argument(
		'--country', '-c',
		help='Console\'s effective country',
		type=regionaldata.Country.get_country
	)
	gen_parser.add_argument(
		'--language', '-l',
		help='Console\'s effective language',
		type=regionaldata.Language.get_language
	)
	gen_parser.add_argument(
		'--out', '-o',
		help='Output generated json file, printed to stdout if not given',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	gen_parser.set_defaults(func=_gen_json)

	check_parser = subparsers.add_parser('CheckReg', help='Check registry status')
	check_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	check_parser.set_defaults(func=_check_register)

	move_parser = subparsers.add_parser('SysTransfer', help='Systransfer account across two consoles (No NNID moved)')
	move_parser.add_argument(
		'--source', '-s',
		help='Console JSON with account to send over',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	move_parser.add_argument(
		'--target', '-t',
		help='Console JSON to receive the account',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	move_parser.set_defaults(func=_move_account)

	move_nnid_parser = subparsers.add_parser('NNIDTransfer', help='Transfer NNID only across two consoles (Use with caution)')
	move_nnid_parser.add_argument(
		'--source', '-s',
		help='Console JSON with account to send over',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	move_nnid_parser.add_argument(
		'--target', '-t',
		help='Console JSON to receive the NNID',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	move_nnid_parser.set_defaults(func=_move_nnid)

	move_check_parser = subparsers.add_parser('LastTransfer', help='Check information about last transfer')
	move_check_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	move_check_parser.set_defaults(func=_move_check_account)

	givs_parser = subparsers.add_parser('GetIVS', help='Get IVS from servers if existant')
	givs_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	givs_parser.set_defaults(func=_get_ivs)

	sivs_parser = subparsers.add_parser('SetIVS', help='Set IVS to servers')
	sivs_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	sivs_parser.set_defaults(func=_set_ivs)

	rivs_parser = subparsers.add_parser('RecoverIVS', help='Attempt recover movable.sed from otp only')
	rivs_parser.add_argument(
		'--otp',
		help='Console\'s encrypted or decrypted otp file',
		required=True,
		type=argparse.FileType('rb')
	)
	rivs_parser.add_argument(
		'--region', '-r',
		help='Console\'s effective region',
		required=True,
		type=regionaldata.Region.get_region
	)
	rivs_parser.add_argument(
		'--country', '-c',
		help='Console\'s effective country',
		required=True,
		type=regionaldata.Country.get_country
	)
	rivs_parser.add_argument(
		'--language', '-l',
		help='Console\'s effective language',
		type=regionaldata.Language.get_language
	)
	rivs_parser.add_argument(
		'--out', '-o',
		help='Output recovered movable.sed',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	rivs_parser.set_defaults(func=_recover_ivs_otp)

	edel_parser = subparsers.add_parser('EShopDelete', help='Delete EShop account')
	edel_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	edel_parser.set_defaults(func=_del_eshop)

	erchange_parser = subparsers.add_parser('EShopRegionChange', help='Attempt to change EShop account region')
	erchange_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	erchange_parser.add_argument(
		'--region', '-r',
		help='Console\'s effective region',
		required=True,
		type=regionaldata.Region.get_region_str
	)
	erchange_parser.add_argument(
		'--country', '-c',
		help='Console\'s effective country',
		required=True,
		type=regionaldata.Country.get_country_str
	)
	erchange_parser.add_argument(
		'--language', '-l',
		help='Console\'s effective language',
		type=regionaldata.Language.get_language_str
	)
	erchange_parser.set_defaults(func=_eshop_region_change)

	etiks_parser = subparsers.add_parser('ETickets', help='List owned ETickets')
	etiks_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	etiks_parser.add_argument(
		'--out', '-o',
		help='Optional output path',
		type=lambda x: pathlib.Path(x).resolve()
	)
	etiks_parser.add_argument(
		'--format', '-f',
		help='Optional output format (text/json)',
		type=str,
		default='text'
	)
	etiks_parser.set_defaults(func=_list_etickets)

	etikdown_parser = subparsers.add_parser('ETicketDownload', help='Download owned ETickets')
	etikdown_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	etikdown_parser.add_argument(
		'--out', '-o',
		help='Output folder path',
		required=True,
		type=lambda x: pathlib.Path(x).resolve()
	)
	etikdown_parser.add_argument(
		'--decrypt', '-d',
		help='Dump ticket with key decrypted from console-unique layer. Will corrupt if imported with AM',
		action='store_true',
		default=False
	)
	etikdown_select = etikdown_parser.add_mutually_exclusive_group(required=True)
	etikdown_select.add_argument(
		'--all', '-a',
		help='Download all tickets',
		action='store_true',
		default=False
	)
	etikdown_select.add_argument(
		'--ticket-id', '-t',
		help='Ticket ids to download, invalid ones will be skipped',
		type=str,
		nargs='+',
		action='append'
	)
	etikdown_parser.set_defaults(func=_download_etickets)

	# TODO:
	#  - allow downloading contents to folder
	#  - ability to generalize tickets to not be console specific
	etitledown_parser = subparsers.add_parser('ETikTitleDownload', help='Download title of owned ETickets, currently only installable on target console with FBI or similar tool, not Godmode9 or custom-installer.')
	etitledown_parser.add_argument(
		'--console', '-C',
		help='Console\'s JSON',
		required=True,
		type=lambda x: str(pathlib.Path(x).resolve())
	)
	etitledown_parser.add_argument(
		'--out', '-o',
		help='Output folder path',
		required=True,
		type=lambda x: pathlib.Path(x).resolve()
	)
	etitledown_parser.add_argument(
		'--ignore-rights', '-i',
		help='Ignore digital right checks',
		action='store_true',
		default=False
	)
	etitledown_parser.add_argument(
		'--keep-going', '-k',
		help='For DLC, keep downloading rest of optional content if another optional is missing',
		action='store_true',
		default=False
	)
	etitledown_select = etitledown_parser.add_mutually_exclusive_group(required=True)
	etitledown_select.add_argument(
		'--all', '-a',
		help='Download all owned titles',
		action='store_true',
		default=False
	)
	etitledown_select.add_argument(
		'--ticket-id', '-t',
		help='Ticket ids to download, invalid ones will be skipped',
		type=str,
		nargs='+',
		action='append'
	)
	etitledown_parser.set_defaults(func=_download_title_tikid)
	args = parser.parse_args(args)

	if args.boot9:
		keys.register_additional_b9_paths(args.boot9)

	if not keys._load_b9_keys():
		print("There was a problem loading boot9 keys, some operations may fail.")

	args.func(parser, args)



from struct import pack, unpack
import typing, re

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, padding as crypto_padding
from cryptography.hazmat.primitives import hashes

from .constants import _load_cfg_rsa_keys, _cfg_pubnums
from . import regionaldata
from .exception import ClassInitError

__all__ = [
	"SecureInfo"
]

_load_cfg_rsa_keys()

class SecureInfo:
	def __init__(self, data: typing.SupportsBytes, is_dev: typing.Optional[bool] = None):
		self._secinfo_data = bytes(data)[:0x111]

		try:
			segments = unpack(">256sBB15s", self._secinfo_data)
		except Exception as e:
			raise ClassInitError("Unpack error") from e

		self._upper_data = self._secinfo_data[0x100:]

		self._sign_block = segments[0]
		self._region = segments[1]
		self._unk_0x101 = segments[2]
		self._serial = segments[3].decode('ascii','replace').split('\x00', 1)[0]

		if not len(self._serial):
			raise ClassInitError("No serial number")

		if not self.validate_serial(self._serial):
			raise ClassInitError("Serial number could not be validated")

		self._is_signed = False
		self._is_dev = False

		if is_dev is not None:
			self._is_dev = bool(is_dev)

			try:
				key = _cfg_pubnums['SECINFO_DEV'] if is_dev else _cfg_pubnums['SECINFO_RETAIL']

				if key is None:
					_load_cfg_rsa_keys()
					key = _cfg_pubnums['SECINFO_DEV'] if is_dev else _cfg_pubnums['SECINFO_RETAIL']

				key = key.public_key(default_backend())
				key.verify(self._sign_block, self._upper_data, crypto_padding.PKCS1v15(), hashes.SHA256())
				self._is_signed = True
			except Exception:
				pass
		else:
			for i, key in enumerate(['SECINFO_RETAIL', 'SECINFO_DEV']):
				try:
					key = _cfg_pubnums[key]

					if key is None:
						_load_cfg_rsa_keys()
						key = _cfg_pubnums[key]

					key = key.public_key(default_backend())
					key.verify(self._sign_block, self._upper_data, crypto_padding.PKCS1v15(), hashes.SHA256())
					self._is_signed = True
					self._is_dev = bool(i)
					break
				except Exception:
					pass

	def __bytes__(self):
		return self._secinfo_data

	def __len__(self):
		return len(self._secinfo_data)

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		if not self.is_signed:
			return f"<{modname}.{clsname} is_signed={self.is_signed} region_str={self.region_str} unk_0x101={self.unk_0x101} serial={self.serial}>"
		else:
			return f"<{modname}.{clsname} is_signed={self.is_signed} is_dev={self.is_dev} region_str={self.region_str} unk_0x101={self.unk_0x101} serial={self.serial}>"

	@staticmethod
	def validate_serial(serial: str) -> typing.Optional[str]:
		try:
			match = re.fullmatch(r'^((?:[A-Z]{2,3})([0-9]{8}))([0-9])?$', serial.upper())
			if not match:
				return None

			sn = match.group(1)
			digits = [int(i) for i in match.group(2)]
			check_digit = match.group(3)
			check_digit = int(check_digit) if check_digit is not None else None
		except:
			return None

		if check_digit is None:
			return sn

		odd_group_sum = sum(digits[::2])
		even_group_sum = sum(digits[1::2])

		calc_step = ((3 * even_group_sum) + odd_group_sum) % 10
		calc_digit = (10 - calc_step) if calc_step != 0 else 0

		if check_digit != calc_digit:
			return None

		return sn

	@classmethod
	def create_from_serial_number(
		cls,
		serial: str,
		region: regionaldata.RegionType,
		is_dev: typing.Optional[bool] = None
	) -> typing.Optional['SecureInfo']:
		serial = cls.validate_serial(str(serial))
		region = regionaldata.Region.get_region(region)
		if not serial or region is None:
			return None

		return cls(pack('>256xBx15s', int(region), serial.encode()), is_dev)

	@property
	def region(self) -> int:
		return self._region
	
	@property
	def unk_0x101(self) -> int:
		return self._unk_0x101

	@property
	def serial(self) -> str:
		return self._serial

	@property
	def region_str(self) -> typing.Union[str, None]:
		return regionaldata.Region.get_region_str(self._region)

	@property
	def is_signed(self) -> bool:
		"""Check this before is_dev or is_retail, cannot guarantee either if not signed"""
		return self._is_signed

	@property
	def is_dev(self) -> bool:
		return self._is_dev

	@property
	def is_retail(self) -> bool:
		return not self._is_dev



# a more quickly put together wrapper for nnid/act stuff
# Thanks zeroskill for extra information on ACT

import typing, base64
import defusedxml.ElementTree as ET
from ..xml_helper import XmlParseHelper, XMLParseError, XML_Element

from .simpledevice import SimpleCtrDevice
from ..connection import Connection, SimpleDownloadBuffer
from .title import Title, MediaType
from .secureinfo import SecureInfo
from . import regionaldata
from .exception import ClassInitError, DataProcessingError, CTRExceptionBase
from .ssl import _ssl_certs
from .constants import _load_act_secrets, _act_cli_data
from ._py_ver_fixes import CTRModel_T

__all__ = [
	"ActSimpleObj",
	"ActEmptyResponse",
	"ActIgnoredContentResponse",
	"ActErrorElement",
	"CommunicationError",
	"XMLError"
]

_load_act_secrets()

class CommunicationError(CTRExceptionBase):
	"""General communication exception"""

class XMLError(CTRExceptionBase):
	"""General xml exception"""

_base_url = 'https://account.nintendo.net/v1/api'

_nver_tids = (
	(
		0x000400db00016202,
		0x000400db00016302,
		0x000400db00016102,
		None,
		0x000400db00016402,
		0x000400db00016502,
		0x000400db00016602
	),
	(
		0x000400db20016202,
		0x000400db20016302,
		0x000400db20016102,
		None,
		None,
		0x000400db20016502,
		None
	)
)

class ActErrorElement:
	cause: typing.Optional[str] = None
	code: typing.Optional[int] = None
	message: typing.Optional[str] = None

	@property
	def is_malformed(self) -> bool:
		return self.code is None or self.message is None

	@staticmethod
	def _xml_get_error_element(
		parent: XmlParseHelper,
		element: XML_Element
	) -> 'ActErrorElement':
		error = ActErrorElement()

		# code and message are not optional for a correct error message, act ignores cause however
		error.cause = parent._xml_element_parse(element, 'cause', parent._xml_get_str_element, True)
		error.code = parent._xml_element_parse(element, 'code', parent._xml_get_int_element, True)
		error.message = parent._xml_element_parse(element, 'message', parent._xml_get_str_element, True)

		return error

class ActResponseBase(XmlParseHelper):
	def __init__(self, xml_string: typing.Union[str, bytes]):
		self._xml_response = xml_string

		if not len(xml_string):
			self._xml_root = None
			self._errors = tuple()
			self._had_errors = False
			return

		try:
			self._xml_root = ET.fromstring(xml_string)

			if self._xml_root.tag != 'errors':
				self._errors = tuple()
				self._had_errors = False
				return

			self._had_errors = True

			error_list = self._xml_multi_element_parse(self._xml_root, 'error', ActErrorElement._xml_get_error_element, True)
			self._errors = tuple(error_list) if error_list else tuple()
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def errors(self) -> typing.Iterable[ActErrorElement]:
		return self._errors

	@property
	def had_errors(self) -> bool:
		return self._had_errors

class ActEmptyResponse(ActResponseBase):
	def __init__(self, xml_string: typing.Union[str, bytes]):
		super().__init__(xml_string)

		if self._xml_root:
			raise DataProcessingError("Unexpected response data!")

class ActIgnoredContentResponse(ActResponseBase):
	def __init__(self, xml_string: typing.Union[str, bytes]):
		super().__init__(xml_string)

class ActSimpleObj:
	def __init__(
		self,
		device: SimpleCtrDevice,
		title: Title
	):
		if not isinstance(device, SimpleCtrDevice):
			raise ClassInitError("Expected SimpleCtrDevice for ActSimpleObj")

		if not isinstance(title, Title):
			raise ClassInitError("Expected title to be Title object")

		# at least a few cases where i'd not expect title ids from in normal scenarios
		if not Title.is_ctr(title) or Title.is_ctrtwl(title) \
			or Title.is_ctrdlc(title) or Title.is_ctrupdate(title):
			raise ClassInitError("Unexpected title id")

		self._device = device
		region = regionaldata.Region.get_region(device.region)
		if region is None:
			raise ClassInitError("Invalid region")

		nver_tid = _nver_tids[device.is_n3ds][region]
		if nver_tid is None:
			raise ClassInitError("Invalid region")

		ntitles = device.titles(MediaType.NAND)
		i = ntitles.index(nver_tid)
		if i < 0:
			raise ClassInitError("NVER was not found on device object")

		self._nver = ntitles[i]
		self._title = title

	def _model_select(self) -> CTRModel_T:
		model = self._device.model_override
		if model is not None:
			return model

		# CFG handles this some way, but I have not seen 100% how
		# one thing is clear, returns some value that ACT then converts to these things
		# if value is not in range, does 3DS-%u format into a buffer
		# but I dont know what makes CFG return the values that it does asides maybe using SN as reference
		model_names = {
			0: 'CTR',
			1: 'SPR',
			2: 'FTR',
			3: 'KTR',
			4: 'RED',
			5: 'JAN'
		}
		retail_sn_firstletters = ['C', 'S', 'A', 'Y', 'Q', 'N']
		dev_sn_firstletters = ['E', 'R', 'P', 'Y', 'Q', 'N']
		_s = dev_sn_firstletters if self._device.is_dev else retail_sn_firstletters
		try:
			i = _s.index(self._device.serial_no[0])
		except ValueError:
			i = 3 if self._device.is_n3ds else 0
		return model_names.get(i, 'CTR')

	def _create_connection(self, url: str):
		_id = _act_cli_data['CLIENT_ID']
		_secret = _act_cli_data['CLIENT_SECRET']
		if _id is None and _secret is None:
			_load_act_secrets()
			_id = _act_cli_data['CLIENT_ID']
			_secret = _act_cli_data['CLIENT_SECRET']
			if _id is None and _secret is None:
				raise DataProcessingError("Missing ACT secrets!")

		conn = Connection()
		conn.set_url(url)
		conn.set_cainfo(_ssl_certs._ca_id_path(3))
		conn.set_cli_cert(*_ssl_certs._client_cert_path_tuple(self._device.is_dev))
		conn.set_header('X-Nintendo-Platform-ID', '0')
		conn.set_header('X-Nintendo-Device-Type', '1' if self._device.is_dev else '2')
		conn.set_header('X-Nintendo-Device-ID', f'{self._device.device_id}')
		conn.set_header('X-Nintendo-Serial-Number', f'{self._device.serial_no}')
		conn.set_header('X-Nintendo-System-Version', f'{self._nver.version:04X}')
		region = regionaldata.Region.get_region(self._device.region)
		region = 0 if region is None else (1 << int(region))
		conn.set_header('X-Nintendo-Region', f'{region}')
		conn.set_header('X-Nintendo-Country', self._device.country)
		conn.set_header('Accept-Language', self._device.language)
		conn.set_header('X-Nintendo-Client-ID', _id)
		conn.set_header('X-Nintendo-Client-Secret', _secret)
		conn.set_header('Accept', '*/*')
		conn.set_header('X-Nintendo-API-Version', '0100')
		conn.set_header('X-Nintendo-FPD-Version', '0000')
		conn.set_header('X-Nintendo-Environment', 'D1' if self._device.is_dev else 'L1')
		conn.set_header('X-Nintendo-Title-ID', f'{self._title.id:016X}')
		conn.set_header('X-Nintendo-Unique-ID', f'{((self._title.id >> 8) & 0xFFFFF):05X}')
		conn.set_header('X-Nintendo-Application-Version', f'{self._title.version_major:04X}') # its actually remaster version, but my only good guess without downloading ncch is this
		conn.set_header('X-Nintendo-Device-Model', self._model_select())
		conn.set_header('X-Nintendo-Device-Cert', base64.b64encode(self._device.ct_cert).decode('utf-8'))

		return conn

	def devices_current_migrations_commit(self) -> typing.Tuple[int, ActEmptyResponse]:
		buf = SimpleDownloadBuffer()

		conn = self._create_connection(_base_url + '/devices/@current/migrations/commit')
		conn.set_header('Content-type', 'application/xml')
		conn.set_customrequest('POST')
		conn.set_write_function(buf.write)

		ret = conn.perform()

		res = ActEmptyResponse(buf.get())
		return (ret, res)

	# act receives an operator, here I just hardset ReserveTransfer
	def devices_current_migrations(
		self,
		*,
		delete: bool = True,
		serialnumber: typing.Optional[str] = None, # target, non optional if delete is false
		deviceid: typing.Optional[typing.SupportsInt] = None # target, non optional if delete is false
	) -> typing.Tuple[int, typing.Union[ActEmptyResponse, ActIgnoredContentResponse]]:
		if not delete:
			serialnumber = SecureInfo.validate_serial(serialnumber)
			deviceid = 0 if deviceid is None else int(deviceid)

			if not serialnumber:
				raise DataProcessingError("Invalid serial number")

			if not 0 < deviceid < 0xFFFFFFFF:
				raise DataProcessingError("Invalid device id")

			msg  =  '<?xml version="1.0" encoding="UTF-8"?>'
			msg +=  '<person_migration><new_platform_id>0</new_platform_id>'
			msg += f'<new_serial_number>{serialnumber}</new_serial_number>'
			msg += f'<new_device_id>{deviceid}</new_device_id>'
			msg +=  '<operator>ReserveTransfer</operator></person_migration>'

		buf = SimpleDownloadBuffer()

		conn = self._create_connection(_base_url + '/devices/@current/migrations')

		if delete:
			conn.set_customrequest('DELETE')
		else:
			conn.set_header('Content-type', 'application/xml')
			conn.set_post_data(msg)

		conn.set_write_function(buf.write)

		ret = conn.perform()

		if delete:
			res = ActIgnoredContentResponse(buf.get())
		else:
			res = ActEmptyResponse(buf.get())
		return (ret, res)



import typing, abc
from ..connection import Connection

class SoapSessionManager(metaclass=abc.ABCMeta):
	@abc.abstractmethod
	def sleep_with_rng_curve(
		self,
		sleep_min_time: typing.Union[int, float],
		rng_multiplier: typing.Union[int, float] = 1
	) -> None:
		"""
		For time padding emulation assistance.
		"""

	@abc.abstractmethod
	def get_connection(self) -> Connection:
		"""
		Return either a new or existing managed connection object.
		"""

	@abc.abstractmethod
	def set_service_standby(self, status: bool) -> None:
		"""
		Set service standby status from SOAP envelope response.
		"""

	@property
	@abc.abstractmethod
	def is_time_emu_enabled(self) -> bool:
		"""
		Either to do padding of time or not.
		"""

	@property
	@abc.abstractmethod
	def service_standby(self) -> bool:
		"""
		Indicate if service is in standby.
		"""

	@property
	@abc.abstractmethod
	def device_id(self) -> int:
		"""
		Device's uint32 device id.
		"""

	@property
	@abc.abstractmethod
	def message_id_step(self) -> int:
		"""
		Last chunk for <MessageId>, initially should be unique to the last SOAP sent.
		E.g. 3DS uses current boot time in ticks converted to microseconds.
		"""

	@property
	@abc.abstractmethod
	def has_tokens(self) -> bool:
		"""
		Indication of ST/WT tokens being available.
		"""

	@property
	@abc.abstractmethod
	def st_token(self) -> typing.Optional[str]:
		"""
		Returns the ST token given by Identification SOAP server.
		This token is more critical to identify the device.
		"""

	@property
	@abc.abstractmethod
	def wt_token(self) -> typing.Optional[str]:
		"""
		Returns the WT token hashed from ST token.
		It's lower caps hexadecimal string of the MD5 digest of the ST token.
		"""

	@property
	@abc.abstractmethod
	def account_id(self) -> typing.Optional[int]:
		"""
		Account id retrieved from Identification SOAP server.
		"""

	@property
	@abc.abstractmethod
	def account_status(self) -> typing.Optional[str]:
		"""
		Account status indicator.
		"""

	@property
	@abc.abstractmethod
	def application_id(self) -> typing.Optional[int]:
		"""
		uint64 application title id that is requesting the service to do SOAPs.
		"""

	@property
	@abc.abstractmethod
	def tin(self) -> typing.Optional[int]:
		"""
		Application's TIN.
		"""

	@property
	@abc.abstractmethod
	def age(self) -> typing.Optional[int]:
		"""
		Some positive number no bigger than 255.
		Still not sure of what's its effect.
		Session only value, not stored.
		"""

	@property
	@abc.abstractmethod
	def region(self) -> typing.Optional[str]:
		"""
		Region short string.
		"""

	@property
	@abc.abstractmethod
	def country(self) -> typing.Optional[str]:
		"""
		Country short string.
		"""

	@property
	@abc.abstractmethod
	def language(self) -> typing.Optional[str]:
		"""
		Language short string.
		"""

	@property
	@abc.abstractmethod
	def serial_no(self) -> str:
		"""
		Console's serial number string.
		"""

	@property
	@abc.abstractmethod
	def user_agent(self) -> str:
		"""
		User-agent string used for http(s) communications.
		"""

	@property
	@abc.abstractmethod
	def keepalive(self) -> bool:
		"""
		Specify specifically connection header keep-alive or close.
		"""

	@property
	@abc.abstractmethod
	def ssl_cert_path(self) -> typing.Optional[str]:
		"""
		SSL certificates used to verify https.
		"""

	@property
	@abc.abstractmethod
	def ssl_cli_cert_paths(self) -> typing.Optional[typing.Tuple[str, ...]]:
		"""
		SSL client certificates for client authentification.
		"""



# This file is originally a part of pyctr.
#
# Copyright (c) 2017-2021 Ian Burgwin
# This file is licensed under The MIT License (MIT).
# A copy of license as been placed here:

## -- LICENSE --
# The MIT License (MIT)
#
# Copyright (c) 2017-2021 Ian Burgwin
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
## -- LICENSE END --

# This is lent from pyctr, from pyctr/util.py @ 854af75, in part, not the full file and some edits.
# Mainly to work independently from pyctr and keep a consistency with pyctr used paths at least.

import typing, os
from sys import platform

__all__ = ['windows', 'macos', 'config_dirs']

windows: bool = platform == 'win32'
macos: bool = platform == 'darwin'

_home = os.path.expanduser('~')
config_dirs: typing.List[str] = [os.path.join(_home, '.3ds'), os.path.join(_home, '3ds')]
if windows:
	config_dirs.insert(0, os.path.join(os.environ.get('APPDATA'), '3ds'))
elif macos:
	config_dirs.insert(0, os.path.join(_home, 'Library', 'Application Support', '3ds'))



import urllib.parse, json, typing

from .soap.manager import CtrSoapManager
from . import exception
from .ssl import _ssl_certs
from ..connection import Connection, SimpleDownloadBuffer

class InvalidManagerStateException(exception.CTRExceptionBase):
	"""General Invalid State of the Soap Manager error"""

class NinjaException(exception.CTRExceptionBase):
	"""General Ninja response error"""
	def __init__(self, *args, **kwds):
		errorcode = kwds.pop('errorcode', -1)
		errormessage = kwds.pop('errormessage', None)
		data = kwds.pop('data', None)

		super().__init__(*args, **kwds)

		self._errorcode = errorcode
		self._errormessage = errormessage
		self._data = data

	@property
	def errorcode(self) -> typing.Optional[int]:
		return self._errorcode

	@property
	def errormessage(self) -> typing.Optional[str]:
		return self._errormessage

	@property
	def responsedata(self) -> typing.Optional[bytes]:
		return self._data

# Very very bare object, early object dev
class NinjaManager:
	def __init__(self, soap_manager: CtrSoapManager):
		if not isinstance(soap_manager, CtrSoapManager):
			raise exception.ClassInitError("Excepted a CTR SOAP Manager")
		self.soap_manager = soap_manager
		self.user_agent = 'CARDBOARD-6.0' # there's information i still need to collect to do a TIGER request
		self.is_cardboard = True

	def open_without_nna(self) -> dict:
		if self.soap_manager.need_register:
			return InvalidManagerStateException("Soap manager reports unregistered state, please register or check register status.")
		# if Cardboard, TIGER has more paramaters
		params = urllib.parse.urlencode(
			{
				'device_token': self.soap_manager.st_token,
				'device_account': self.soap_manager.account_id,
				'device_id': self.soap_manager.device_id,
				'lang': self.soap_manager.language,
				'country': self.soap_manager.country,
				'serial_number': self.soap_manager.serial_no
			}
		)

		buf = SimpleDownloadBuffer()

		conn = Connection()
		conn.set_url('https://ninja.ctr.shop.nintendo.net/ninja/ws/my/session/!open_without_nna')
		conn.set_keepalive(False)
		conn.set_header("User-Agent", self.user_agent)
		conn.set_header("Accept", "application/json")
		conn.set_header("Accept", "application/x-www-form-urlencoded") # cardboard sets this too, because
		conn.set_header("Content-Type", "application/x-www-form-urlencoded")
		conn.set_cainfo(_ssl_certs._ca_id_path(3))
		conn.set_cli_cert(*self.soap_manager.ssl_cli_cert_paths)
		conn.set_post_data(params.encode('ascii'))
		conn.set_write_function(buf.write)
		ret = conn.perform()

		if ret != 200:
			try:
				parsed = json.loads(buf.get().decode('utf-8'))
				error = parsed.get('error', None)
				errorcode = int(error.get('code', -1))
				errormessage = error.get('message', None)
			except:
				errorcode = -1
				errormessage = None
			raise NinjaException(
				"Non 200 http response",
				errorcode=errorcode,
				errormessage=errormessage,
				data=buf.get()
			)

		try:
			return json.loads(buf.get().decode('utf-8'))
		except:
			raise NinjaException("Parse error.", data=buf.get())



import hashlib, typing, base64
from enum import IntEnum, unique

from ...nintendowifi.soapsessionmanager import SoapSessionManager
from ..simpledevice import SimpleCtrDevice
from ..title import Title, MediaType
from .. import exception
from .. import regionaldata
from ..ssl import _ssl_certs
from ...connection import Connection
from ..otp import CTCert

__all__ = [
	"ServiceLevel",
	"CtrSoapManager"
]

@unique
class ServiceLevel(IntEnum):
	SHOP = 0,
	SYSTEM = 1

class CtrSoapManager:
	def __init__(self, device: SimpleCtrDevice, nup_agent: bool):
		if not isinstance(device, SimpleCtrDevice):
			raise exception.ClassInitError("Device should be a SimpleCtrDevice")

		self._useragent = f'CTR {"NUP" if nup_agent else "EC"} 040600 Mar 14 2012 13:32:39'
		self._keepalive = True
		self._device = device
		self._appid = None
		self._tin = None
		self._service_urls = {}
		self._service_level = ServiceLevel.SYSTEM if nup_agent else ServiceLevel.SHOP
		self._service_standby = False
		self._ivs_sync_flag = False

		self._connection = Connection()

		self._service_urls['ecs'] = 'https://ecs.c.shop.nintendowifi.net/ecs/services/ECommerceSOAP'
		self._service_urls['ias'] = 'https://ias.c.shop.nintendowifi.net/ias/services/IdentityAuthenticationSOAP'
		self._service_urls['cas'] = 'https://cas.c.shop.nintendowifi.net/cas/services/CatalogingSOAP'
		self._service_urls['nus'] = 'https://nus.c.shop.nintendowifi.net/nus/services/NetUpdateSOAP'

		self._overrides = {}

	def set_app_and_tin(self, appid: typing.Optional[typing.SupportsInt] = None, tin: typing.Optional[typing.SupportsInt] = None):
		if appid is not None:
			appid = int(appid)

		if tin is not None:
			tin = int(tin)

		if (appid and appid < 0) or (tin and tin < 0):
			return exception.DataProcessingError("Expecting positive appid and tin")

		self._appid = appid
		self._tin = tin

	def sleep_with_rng_curve(
		self,
		sleep_min_time: typing.Union[int, float],
		rng_multiplier: typing.Union[int, float] = 1
	) -> None:
		self._device.sleep_with_rng_curve(sleep_min_time, rng_multiplier)

	def get_connection(self) -> Connection:
		return self._connection

	def set_keepalive(self, toggle: bool) -> None:
		self._keepalive = bool(toggle)

	def set_service_standby(self, status: bool) -> None:
		self._service_standby = bool(status)

	def set_ivs_sync_flag(self, status: bool) -> None:
		self._ivs_sync_flag = bool(status)

	def register_account(
		self,
		status: str,
		st_token: str,
		account_id: int,
		country: str,
		ext_account_id: typing.Optional[str] = None,
		device_code: typing.Optional[str] = None,
		register_expired: bool = False
	) -> None:
		if self._service_level == ServiceLevel.SYSTEM:
			return
		if not isinstance(status, str) or not isinstance(st_token, str) or not isinstance(account_id, int) or not isinstance(country, str):
			raise exception.DataProcessingError("Invalid type for status, st_token, account_id or country")
		if status == 'U':
			raise exception.DataProcessingError("Can't register with an unregister, use unregister_account instead")
		last_ivs = self._device._ecommerce_info.get('last_ivs', None) if self.account_id is not None and self.account_id == account_id else None
		ecommerce_info = {
			'st_token': st_token,
			'wt_token': hashlib.md5(st_token.encode('ascii')).hexdigest(),
			'account_id': account_id,
			'account_status': status,
			'register_expired': register_expired,
			'last_ivs': last_ivs
		}
		if ext_account_id:
			ecommerce_info['ext_account_id'] = ext_account_id
		if device_code:
			ecommerce_info['device_code'] = device_code
		self.country_change(country)
		self.drop_account()
		self._device.update_ecommerce(ecommerce_info)

	def unregister_account(self) -> None:
		if self._service_level == ServiceLevel.SYSTEM:
			return
		self._device.update_ecommerce({'account_status': 'U'})

	def drop_account(self) -> None:
		if self._service_level == ServiceLevel.SYSTEM:
			return
		self._device.drop_ecommerce()

	def ap_sign(self, title_id: typing.SupportsInt, data: typing.SupportsBytes) -> typing.Tuple[bytes, bytes]:
		return self._device.ap_sign(title_id, data)

	def region_change(
		self,
		region: regionaldata.RegionType,
		country: typing.Optional[regionaldata.CountryType] = None,
		language: typing.Optional[regionaldata.LanguageType] = None
	) -> None:
		self._device.region_change(region, country, language)
		self.clear_all_overrides()
		self.drop_account()

	def country_change(
		self,
		country: regionaldata.CountryType
	) -> None:
		region = regionaldata.Country.get_region_by_country(country)
		if region != regionaldata.Region.get_region(self._device.region):
			self.region_change(region, country)
		else:
			self._device.country_change(country)
		self.clear_country_override()

	def language_change(
		self,
		language: typing.Optional[regionaldata.LanguageType] = None
	) -> None:
		self._device.language_change(language)
		self.clear_language_override()

	def set_region_override(self, region: regionaldata.RegionType) -> None:
		_region = regionaldata.Region.get_region_str(region)
		if _region is None:
			raise exception.DataProcessingError("Invalid region override")
		self._overrides['region'] = _region

	def set_country_override(self, country: regionaldata.CountryType) -> None:
		_country = regionaldata.Country.get_country_str(country)
		if _country is None:
			raise exception.DataProcessingError("Invalid country override")
		self._overrides['country'] = _country

	def set_language_override(self, language: typing.Optional[regionaldata.LanguageType]) -> None:
		_language = regionaldata.Language.get_language_str(language)
		if _language is None and language is not None:
			raise exception.DataProcessingError("Invalid language override")
		self._overrides['language'] = _language

	def clear_region_override(self) -> None:
		if 'region' in self._overrides.keys():
			del self._overrides['region']

	def clear_country_override(self) -> None:
		if 'country' in self._overrides.keys():
			del self._overrides['country']

	def clear_language_override(self) -> None:
		if 'language' in self._overrides.keys():
			del self._overrides['language']

	def clear_all_overrides(self) -> None:
		self._overrides.clear()

	def get_overrides_copy(self) -> typing.Dict[str, typing.Optional[str]]:
		return self._overrides.copy()

	def load_overrides_copy(self, overrides: typing.Dict[str, typing.Optional[str]]) -> None:
		o = self.get_overrides_copy()
		try:
			region = overrides.get('region', None)
			country = overrides.get('country', None)
			language = overrides.get('language', None)
			if region is not None:
				self.set_region_override(region)
			else:
				self.clear_region_override()
			if country is not None:
				self.set_country_override(country)
			else:
				self.clear_country_override()
			if 'language' in overrides.keys():
				self.set_language_override(language)
			else:
				self.clear_language_override()
		except:
			self._overrides = o
			raise

	def commit_overrides(self) -> None:
		_region = self._overrides.get('region', None)
		_country = self._overrides.get('country', None)
		_language = self._overrides.get('language', None)

		if _region is None and _country is None and _language is None:
			self.clear_all_overrides()
			return

		# country takes priority, as wrong country settings cause registers on wrong regions on real SOAP register
		if _country is not None:
			self.country_change(_country)
		elif _region is not None:
			self.region_change(_region)

		if _language is not None:
			self.language_change(_language)

		self.clear_all_overrides()

	def is_service_ready(self) -> bool:
		if self._service_level == ServiceLevel.SHOP:
			if self._service_urls.get('ecs', None) and \
				self._service_urls.get('cas', None) and \
				self._service_urls.get('content_prefix', None) and \
				self._service_urls.get('uncached_content_prefix', None) and \
				(self._service_urls.get('ias', None) or self._service_standby):
				return True
		else:
			if self._service_urls.get('nus', None) and \
				self._service_urls.get('system_content_prefix', None) and \
				self._service_urls.get('system_uncached_content_prefix', None):
				return True
		return False

	def get_url_by_identifier(self, name: str) -> typing.Optional[str]:
		return self._service_urls.get(name, None)

	def set_url_by_identifier(self, name: str, url: str) -> None:
		if not isinstance(name, str) or not isinstance(name, str):
			raise exception.DataProcessingError("Excepting str for name and url")

		self._service_urls[name] = url

	def export_ivs(self) -> bytes:
		return self._device.export_ivs()

	def import_ivs(self, ivs: typing.SupportsBytes) -> None:
		self._device.import_ivs(ivs)

	def set_last_sync_ivs(self, ivs: typing.SupportsBytes) -> None:
		self._device._ecommerce_info['last_ivs'] = base64.b64encode(bytes(ivs)).decode('utf-8')

	def titles(self, media: MediaType) -> typing.Iterable[Title]:
		return self._device.titles(media)

	@property
	def last_sync_ivs(self) -> typing.Optional[bytes]:
		ivs = self._device._ecommerce_info.get('last_ivs', None)
		if not ivs:
			return None
		return base64.b64decode(ivs)

	@property
	def should_query_registry(self) -> bool:
		return not self._device.has_ecommerce_info

	@property
	def need_register(self) -> bool:
		"""
		This should be only used if .should_query_registry says False
		True if == U.
		T and R and P are different, should check with other conditional checks.
		"""
		return self.account_status == 'U'

	@property
	def need_account_transfer(self) -> bool:
		"""
		This should be only used if .should_query_registry says False
		True if == P.
		U and R and T are different, should check with other conditional checks.
		"""
		return self.account_status == 'P'

	@property
	def service_level(self) -> ServiceLevel:
		return self._service_level

	@property
	def is_time_emu_enabled(self) -> bool:
		return self._device.time_padding_emu

	@property
	def service_standby(self) -> bool:
		return self._service_standby

	@property
	def ivs_sync_flag(self) -> bool:
		return self._ivs_sync_flag

	@property
	def is_msed_signed(self) -> bool:
		return self._device.msed.lfcs.is_signed

	@property
	def ec_version(self) -> str:
		return 'EC 4.6.0 Mar 14 2012 13:32:39'

	@property
	def ext_account_id(self) -> typing.Optional[str]:
		return self._device._ecommerce_info.get('ext_account_id', None)

	@property
	def device_code(self) -> typing.Optional[str]:
		return self._device._ecommerce_info.get('device_code', None)

	@property
	def ct_cert(self) -> bytes:
		return self._device.ct_cert

	@property
	def ct_cert_full(self) -> CTCert:
		return self._device.ct_cert_full

	@property
	def device_id(self) -> int:
		return self._device.device_id | (4 << 32)

	@property
	def message_id_step(self) -> int:
		return self._device.boot_time_ms

	@property
	def has_tokens(self) -> bool:
		return True if self.st_token else False

	@property
	def st_token(self) -> typing.Optional[str]:
		return self._device._ecommerce_info.get('st_token', None)

	@property
	def wt_token(self) -> typing.Optional[str]:
		return self._device._ecommerce_info.get(
			'wt_token',
			hashlib.md5(self.st_token.encode('ascii')).hexdigest() if self.has_tokens else None
		)

	@property
	def account_id(self) -> typing.Optional[int]:
		return self._device._ecommerce_info.get('account_id', None)

	@property
	def account_status(self) -> typing.Optional[str]:
		return self._device._ecommerce_info.get('account_status', None)

	@property
	def register_expired(self) -> bool:
		return self._device._ecommerce_info.get('register_expired', False)

	@property
	def application_id(self) -> typing.Optional[int]:
		return self._appid

	@property
	def tin(self) -> typing.Optional[int]:
		return self._tin

	@property
	def age(self) -> typing.Optional[int]:
		return None # stub for now

	@property
	def region(self) -> typing.Optional[str]:
		return self._device.region if 'region' not in self._overrides.keys() else self._overrides['region']

	@property
	def country(self) -> typing.Optional[str]:
		return self._device.country if 'country' not in self._overrides.keys() else self._overrides['country']

	@property
	def language(self) -> typing.Optional[str]:
		return self._device.language if 'language' not in self._overrides.keys() else self._overrides['language']

	@property
	def serial_no(self) -> str:
		return self._device.serial_no

	@property
	def user_agent(self) -> str:
		return self._useragent

	@property
	def keepalive(self) -> bool:
		return self._keepalive

	@property
	def ssl_cert_path(self) -> typing.Optional[str]:
		return _ssl_certs._ca_id_path(3)

	@property
	def ssl_cli_cert_paths(self) -> typing.Optional[typing.Tuple[str, str]]:
		return _ssl_certs._client_cert_path_tuple(self._device.is_dev)

SoapSessionManager.register(CtrSoapManager)



import pathlib, typing

from ..environment import find_in_config_dirs
from .. import keys

_certs_dir = pathlib.Path(__file__).resolve().parent / '_certificates'

_cert_files_by_id = {
	0x1: 'Nintendo CA.pem',
	0x2: 'Nintendo CA - G2.pem',
	0x3: 'Nintendo CA - G3.pem',
	0x4: 'Nintendo Class 2 CA.pem',
	0x5: 'Nintendo Class 2 CA - G2.pem',
	0x6: 'Nintendo Class 2 CA - G3.pem',
	0x7: 'GTE CyberTrust Global Root.pem',
	0x8: 'AddTrust External CA Root.pem',
	0x9: 'COMODO RSA Certification Authority.pem',
	0xA: 'USERTrust RSA Certification Authority.pem',
	0xB: 'DigiCert High Assurance EV Root CA.pem'
}

def _ca_id_path(_id: int) -> typing.Optional[str]:
	if _id < 0:
		pem = '3ds-ssl-ca-bundle.crt'
	else:
		pem = _cert_files_by_id.get(_id, None)

	if pem is None:
		return None

	path = _certs_dir / pem

	if path.is_file():
		return str(path)

	paths = find_in_config_dirs('SSLCertificates/' + pem)

	return paths[0] if paths else None

def _client_cert_path_tuple(is_dev: bool = False) -> typing.Tuple[typing.Optional[str], typing.Optional[str]]:
	cert = 'ctr-common-1-cert-dev.pem' if is_dev else 'ctr-common-1-cert-prod.pem'
	key = 'ctr-common-1-key-dev.pem' if is_dev else 'ctr-common-1-key-prod.pem'

	certpath = _certs_dir / cert
	keypath = _certs_dir / key

	if not certpath.is_file():
		certpaths = find_in_config_dirs('SSLCertificates/client/' + cert)
		certpath = certpaths[0] if certpaths else None

	if not keypath.is_file():
		keypaths = find_in_config_dirs('SSLCertificates/client/' + key)
		keypath = keypaths[0] if keypaths else None

	return (str(certpath), str(keypath))



import time, random, math, typing

from ._py_ver_fixes import monotonic_ns

__all__ = [
	"BootTimeEmu"
]

def monotonic_µs() -> int:
	return monotonic_ns() // 1000

class BootTimeEmu:
	def __init__(self):
		self.reboot()

	def reboot(self) -> None:
		self._boottime = monotonic_µs()
		self._nextready = self._boottime + random.randint(8000000,12000000)

	def sleep_with_rng_curve(
		self,
		sleep_min_time: typing.Union[int, float],
		rng_multiplier: typing.Union[int, float] = 1
	) -> None:
		# tMin + random() ** 2.5 * rng_multiplier
		timenow = monotonic_µs()
		t = sleep_min_time + math.pow(random.random(), 2.5) * rng_multiplier
		t_F = math.floor(t)
		remaining_last = self._nextready - timenow if self._nextready > timenow else 0
		self._nextready = timenow + t_F * 1000000 + int((t - t_F) * 1000000) + remaining_last

	@property
	def get_microseconds(self) -> int:
		time.sleep(1e-6)

		timenow = monotonic_µs()
		boottime = self._boottime
		nextready = self._nextready

		if nextready > timenow:
			time.sleep((nextready - timenow) / 1000000.0)
			timenow = monotonic_µs()

		return timenow - boottime



from struct import pack, unpack
import typing

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, padding as crypto_padding
from cryptography.hazmat.primitives import hashes

from .constants import _load_cfg_rsa_keys, _cfg_pubnums
from .otp import OTP
from .exception import ClassInitError, DataProcessingError

__all__ = [
	"LFCS"
]

_load_cfg_rsa_keys()

class LFCS:
	def __init__(self, data: typing.SupportsBytes):
		self._lfcs_data = bytes(data)[:0x110]

		if len(self._lfcs_data) != 0x110:
			raise ClassInitError("LFCS data should be 272 bytes")

		segments = unpack("<256s8sQ", self._lfcs_data)

		self._sign_block = segments[0]
		self._upper_data = self._lfcs_data[0x100:]

		if segments[1][0] + sum(segments[1][2:]) != 0:
			raise ClassInitError("LFCS bytes @ 0x100 to 0x107, excluding 0x101, not supported as non 0")

		if 0 != segments[1][1] != 1:
			raise ClassInitError("LFCS byte @ 0x101 expected to be 0 or 1")

		self._is_dev = bool(segments[1][1])
		self._lfcs_id = segments[2]

		try:
			key = _cfg_pubnums['LFCS_DEV'] if self._is_dev else _cfg_pubnums['LFCS_RETAIL']

			if key is None:
				_load_cfg_rsa_keys()
				key = _cfg_pubnums['LFCS_DEV'] if self._is_dev else _cfg_pubnums['LFCS_RETAIL']

			key = key.public_key(default_backend())
			key.verify(self._sign_block, self._upper_data, crypto_padding.PKCS1v15(), hashes.SHA256())
			self._is_signed = True
		except Exception as e:
			self._is_signed = False

	def __bytes__(self):
		return self._lfcs_data

	def __len__(self):
		return len(self._lfcs_data)

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		return f"<{modname}.{clsname} is_signed={self._is_signed!r} is_dev={self._is_dev!r} lfcs_id={self._lfcs_id!r}>"

	@classmethod
	def generate_from_otp(cls, otp: OTP) -> 'LFCS':
		if not isinstance(otp, OTP):
			raise DataProcessingError("Expected OTP")

		lfcs = pack("<256xxB6xQ", 1 if otp.is_dev else 0, otp.lfcs_id)

		return cls(lfcs)

	@property
	def is_signed(self) -> bool:
		return self._is_signed

	@property
	def is_dev(self) -> bool:
		return self._is_dev

	@property
	def lfcs_id(self) -> int:
		return self._lfcs_id



from enum import IntEnum, unique
import typing, re, base64, time
import defusedxml.ElementTree as ET
from ..xml_helper import XmlParseHelper, XMLParseError, XML_Element

from .soapsessionmanager import SoapSessionManager
from ..exception import CleanintyExceptionBase, ClassInitError
from ..connection import Connection, SimpleDownloadBuffer

__all__ = [
	"SoapEnvelopeBase",
	"SoapSubNames",
	"ObjectTimingEmuHelper",
	"InvalidTagError",
	"XMLParseError",
	"SoapError",
	"ManagerTypeError"
]

class InvalidTagError(CleanintyExceptionBase):
	"""General SoapEnvelopeBase invalid tag exception"""

class SoapError(CleanintyExceptionBase):
	"""General SoapEnvelopeBase SOAP error"""

class SoapCodeError(CleanintyExceptionBase):
	"""General SoapEnvelopeBase SOAP ErrorCode error"""
	def __init__(self, *args, **kwds):
		errorcode = kwds.pop('errorcode', -1)
		errormessage = kwds.pop('errormessage', None)

		super().__init__(*args, **kwds)

		self._errorcode = errorcode
		self._errormessage = errormessage

	@property
	def soaperrorcode(self) -> int:
		return self._errorcode

	@property
	def soaperrormessage(self) -> str:
		return self._errormessage

class ManagerTypeError(CleanintyExceptionBase, TypeError):
	"""General SoapEnvelopeBase manager type error"""

_msg_start_template = """<?xml version="1.0" encoding="UTF-8"?>
<SOAP-ENV:Envelope xmlns:SOAP-ENV="http://schemas.xmlsoap.org/soap/envelope/"
                   xmlns:SOAP-ENC="http://schemas.xmlsoap.org/soap/encoding/"
                   xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
                   xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                   xmlns:{0}="urn:{0}.wsapi.broadon.com">
<SOAP-ENV:Body>
<{0}:{1} xsi:type="{0}:{1}RequestType">
<{0}:Version>{2}</{0}:Version>
<{0}:MessageId>EC-{3}-{4}</{0}:MessageId>
"""

_msg_end_template = """</{0}:{1}>
</SOAP-ENV:Body>
</SOAP-ENV:Envelope>
"""

# more intended for __init__ methods, or functions with self and manager as second arg
def ObjectTimingEmuHelper(min_time, rng_multiplier = 0):
	def inner1(func):
		def inner2(self, manager, *args, **kwds):
			if not isinstance(manager, SoapSessionManager):
				raise ManagerTypeError("Can't emulate timing without manager object")
			if not manager.is_time_emu_enabled:
				return func(self, manager, *args, **kwds)
			t = time.monotonic()
			try:
				ret = func(self, manager, *args, **kwds)
			finally:
				t = min_time - (time.monotonic() - t)
				manager.sleep_with_rng_curve(0 if t < 0 else t, rng_multiplier)
			return ret
		return inner2
	return inner1

@unique
class SoapSubNames(IntEnum):
	NUS = 0
	ECS = 1
	IAS = 2
	CAS = 3
	BGS = 4

	@classmethod
	def get_name(cls, value: typing.Union['SoapSubNames', int]) -> typing.Union[str, None]:
		try:
			return cls(int(value)).name.lower()
		except Exception:
			return None

class SoapEnvelopeBase(XmlParseHelper):
	def __init__(
		self,
		sub_name: typing.Union[SoapSubNames, int],
		soap_action_name: str,
		session_manager: SoapSessionManager,
		use_st_token: bool,
		keepalive: bool,
		version2: bool = True
	):
		try:
			sub_name = SoapSubNames(sub_name)
		except Exception as e:
			raise ClassInitError("Invalid Sub Server Name") from e

		if sub_name == SoapSubNames.BGS:
			raise ClassInitError("BGS currently unsupported")

		if not isinstance(session_manager, SoapSessionManager):
			raise ClassInitError("Need to have session manage to be a SoapSessionManager")

		use_st_token = bool(use_st_token)
		version = "2.0" if version2 else "1.0"

		self._sub_name = SoapSubNames.get_name(sub_name)
		self._soap_action_name = str(soap_action_name)
		self._session_manager = session_manager
		self._keepalive = bool(keepalive)

		self._http_headers = []
		self._http_headers.append(['User-Agent', session_manager.user_agent])
		self._http_headers.append(['Connection', 'Keep-Alive' if session_manager.keepalive else 'close'])
		self._http_headers.append(['Accept-Charset', 'UTF-8'])
		self._http_headers.append(['Content-type', 'text/xml; charset=utf-8'])
		self._http_headers.append(['SOAPAction', 'urn:{0}.wsapi.broadon.com/{1}'.format(self._sub_name, self._soap_action_name)])

		self._xml_response_namespaces = {
			'soapenv': 'http://schemas.xmlsoap.org/soap/envelope/',
			'xsd': 'http://www.w3.org/2001/XMLSchema',
			'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
			'urn': 'urn:{0}.wsapi.broadon.com'.format(self._sub_name)
		}

		self._tag_stack = []

		self._response = None

		self._envelope = _msg_start_template.format(
			self._sub_name, # 0
			self._soap_action_name, # 1
			version, # 2
			session_manager.device_id, # 3
			session_manager.message_id_step, # 4
		)

		self._envelope += '<{0}:DeviceId>{1}</{0}:DeviceId>\n'.format(
			self._sub_name,
			session_manager.device_id
		)

		if sub_name != SoapSubNames.NUS:
			if session_manager.has_tokens:
				self._envelope += '<{0}:DeviceToken>{1}-{2}</{0}:DeviceToken>\n'.format(
					self._sub_name,
					"ST" if use_st_token else "WT",
					session_manager.st_token if use_st_token else session_manager.wt_token
				)
			if session_manager.account_id is not None:
				self._envelope += '<{0}:AccountId>{1}</{0}:AccountId>\n'.format(
					self._sub_name,
					session_manager.account_id
				)

		if sub_name == SoapSubNames.ECS or sub_name == SoapSubNames.CAS:
			if session_manager.application_id is not None:
				self._envelope += '<{0}:ApplicationId>{1:016x}</{0}:ApplicationId>\n'.format(
					self._sub_name,
					session_manager.application_id
				)
			if session_manager.tin is not None:
				self._envelope += '<{0}:TIN>{1}</{0}:TIN>\n'.format(
					self._sub_name,
					session_manager.tin
				)

		if session_manager.region is not None:
			self._envelope += '<{0}:{1}>{2}</{0}:{1}>\n'.format(
				self._sub_name,
				"Region" if sub_name != SoapSubNames.NUS else "RegionId",
				session_manager.region
			)

		if session_manager.country is not None:
			self._envelope += '<{0}:{1}>{2}</{0}:{1}>\n'.format(
				self._sub_name,
				"Country" if sub_name != SoapSubNames.NUS else "CountryCode",
				session_manager.country
			)

		if session_manager.language is not None:
			self._envelope += '<{0}:Language>{1}</{0}:Language>\n'.format(
				self._sub_name,
				session_manager.language
			)

		if sub_name == SoapSubNames.ECS or sub_name == SoapSubNames.NUS:
			self._envelope += '<{0}:SerialNo>{1}</{0}:SerialNo>\n'.format(
				self._sub_name,
				session_manager.serial_no
			)

		# something something ServiceHandle, ServiceTicket and ServiceId for ecs
		# something something, they never happen in EC 4.6

		if sub_name == SoapSubNames.CAS and session_manager.age is not None:
			# still not sure what "Age" is in this context
			# also cursed indentation
			self._envelope += '  <{0}:Age>{1}</{0}:Age>\n'.format(
				self._sub_name,
				session_manager.age
			)

		self._closed = False

		self._timestamp = None
		self._errorcode = None
		self._errormessage = None
		self._servicestandbymode = None

	@staticmethod
	def _validate_tagname(name: str) -> bool:
		"""We'll just allow Alphabetic characters"""
		return True if re.fullmatch('^[a-zA-Z]+$', name) else False

	@staticmethod
	def _get_escaped_field(field: str) -> str:
		# there gotta be a better way to do this right?
		# but i guess i can say "it works" for now
		return re.sub('>','&gt;',re.sub('<','&lt;',re.sub("'",'&apos;',re.sub('"','&quot;',re.sub('&','&amp;',field)))))

	def _push_tag(self, name: str) -> None:
		if self._closed:
			return

		name = str(name)
		if not self._validate_tagname(name):
			raise InvalidTagError("Tag invalid for these SOAP purposes")

		self._tag_stack.append(name)

		self._envelope += '<{0}:{1}>'.format(
			self._sub_name,
			name
		)

	def _pop_tag(self) -> None:
		if self._closed:
			return

		self._envelope += '</{0}:{1}>'.format(
			self._sub_name,
			self._tag_stack.pop()
		)

	def _write_tag(self, name: str, field: str) -> None:
		if self._closed:
			return

		field = self._get_escaped_field(str(field))
		self._push_tag(name)
		self._envelope += field
		self._pop_tag()

	def _close(self):
		if self._closed:
			return

		while self._tag_stack:
			self._pop_tag()

		if self._envelope[-1] != '\n':
			self._envelope += '\n'

		self._envelope += _msg_end_template.format(
			self._sub_name, # 0
			self._soap_action_name # 1
		)

		self._closed = True

		# print temporary to view
		# print(self._envelope)

	def _send(self, url: str) -> int:
		self._close()

		_cli_cert = None
		_cli_key = None
		_cli_cert_paths = self._session_manager.ssl_cli_cert_paths
		if _cli_cert_paths:
			_cli_cert = _cli_cert_paths[0]
			if len(_cli_cert_paths) > 1:
				_cli_key = _cli_cert_paths[1]

		buf = SimpleDownloadBuffer()

		conn = self._session_manager.get_connection()
		conn.reset_headers()
		conn.set_url(url)
		conn.set_keepalive(self._keepalive)
		for i in self._http_headers:
			conn.set_header(i[0], i[1])
		conn.set_cainfo(self._session_manager.ssl_cert_path)
		conn.set_cli_cert(_cli_cert, _cli_key)
		conn.set_post_data(self._envelope.encode('utf-8'))
		conn.set_write_function(buf.write)
		ret = conn.perform()

		if ret == 200:
			self._response = buf.get().decode('utf-8')
			#print(self._response)

		return ret

	@staticmethod
	def _xml_get_bool_element(
		parent: XmlParseHelper,
		element: XML_Element
	) -> bool:
		parent._xml_raise_if_any_subelement(element)
		if element.text == '0' or element.text == 'false':
			return False
		if element.text == '1' or element.text == 'true':
			return True
		raise XMLParseError("Invalid boolean field")

	def _initiate_response_parse(
		self
	) -> typing.Tuple[
		XML_Element, # root
		XML_Element # Response element
	]:
		try:
			root = ET.fromstring(self._response)
			body = self._xml_element_parse(root, 'soapenv:Body', self._xml_get_element)
			response = self._xml_element_parse(body, 'urn:{0}Response'.format(self._soap_action_name), self._xml_get_element)
			self._timestamp = self._xml_element_parse(response, 'urn:TimeStamp', self._xml_get_int_element, True)
			self._errorcode = self._xml_element_parse(response, 'urn:ErrorCode', self._xml_get_int_element)
			self._errormessage = self._xml_element_parse(response, 'urn:ErrorMessage', self._xml_get_str_element, True)
			self._servicestandbymode = self._xml_element_parse(response, 'urn:ServiceStandbyMode', self._xml_get_bool_element, self._sub_name != 'ecs')
			if self._sub_name == 'ecs':
				self._session_manager.set_service_standby(self._servicestandbymode)
			return (root, response)
		except CleanintyExceptionBase:
			raise
		except Exception as e:
			raise XMLParseError("Unexpected exception while parsing XML") from e

	@staticmethod
	def _validate_errorcode(errorcode: int, errormessage: str) -> None:
		if errorcode == 0:
			return

		raise SoapCodeError("Soap error code non 0", errorcode=errorcode, errormessage=errormessage)

	@property
	def timestamp(self) -> typing.Optional[int]:
		return self._timestamp

	@property
	def errorcode(self) -> typing.Optional[int]:
		return self._errorcode

	@property
	def errormessage(self) -> typing.Optional[str]:
		return self._errormessage

	@property
	def servicestandbymode(self) -> typing.Optional[bool]:
		return self._servicestandbymode



from struct import pack
import typing, os, pathlib

from .cdn import CDN, CDNDownloadStage, AssistedCDNDownload
from .tmd import ContentChunkRecord, ContentTypeFlags
from .exception import CTRExceptionBase, ClassInitError, DataProcessingError
from ..connection import WriterFuncType, MultiWriter

__all__ = [
	"CiaCDNBuilder",
	"CDNFolderDownload",
	"CDNMultiHelper"
]

def _exception_try_shield(inform_except: str, except_ret: typing.Any = None):
	def inner(func):
		def wrapper(self, *args, **kwargs):
			try:
				return func(self, *args, **kwargs)
			except Exception:
				getattr(self, inform_except)()
				return except_ret
		return wrapper
	return inner

class CiaCDNBuilder:
	def __init__(self, output: str, rollback_optional_failed_content: bool = False):
		self._path = pathlib.Path(output).resolve()
		self._outfile = self._path.open("wb")
		self._errorstate = False
		self._end_flag = False
		self._has_ended = False

		self._cert_chain_size = 0
		self._tik_size = 0
		self._tmd_size = 0
		self._content_off = -1

		self._curr_content_off = -1
		self._limit_write = -1
		self._curr_optional = False
		self._rollback_opt_content = bool(rollback_optional_failed_content)

		self._index_field = 0

		self._last_stage = -1

		self._ensure_seek(0x2040)

	def __del__(self):
		self._end()

	def _block_align_up(self) -> None:
		self._ensure_seek(self._calc_align_up(self._outfile.tell()))

	def _calc_align_up(self, x: int) -> int:
		return ((x + 0x3F) >> 6) << 6

	def _error(self) -> None:
		if self._errorstate:
			return

		if self._rollback_opt_content and \
			self._last_stage == CDNDownloadStage.CONTENT and \
			self._curr_content_off != -1 and \
			self._curr_optional:
			try:
				self._rollback_opt_content = False # safe guard against failed rollbacks
				self._ensure_seek(self._curr_content_off)
				size = self._outfile.truncate()
				if size != self._curr_content_off:
					raise Exception("") # escape
				self._curr_content_off = -1
				self._limit_write = -1
				self._curr_optional = False
				self._rollback_opt_content = True
				return
			except Exception:
				pass

		self._errorstate = True
		try:
			self._outfile.close()
		except Exception:
			pass

		try:
			self._path.unlink()
		except Exception:
			pass

	@_exception_try_shield('_error')
	def _end(self) -> None:
		if self._has_ended:
			return

		if not self._end_flag or \
			not self._tmd_size or \
			self._content_off < 0 or \
			self._curr_content_off >= 0:
			self._error()

		while not self._errorstate:
			end_off = self._outfile.tell()
			aligned_end_off = self._calc_align_up(end_off)
			
			if not self._ensure_seek(aligned_end_off) or \
				self._outfile.truncate() != aligned_end_off:
				self._error()
				break

			self._outfile.seek(0)
			if self._outfile.tell() != 0:
				self._error()
				break

			indexes = int.to_bytes(self._index_field, 0x2000, 'big')

			written = self._outfile.write(
				pack(
					"<I HH IIII Q 8192s",
					0x2020,
					0, 0,
					self._cert_chain_size,
					self._tik_size,
					self._tmd_size,
					0,
					end_off - self._content_off,
					indexes
				)
			)

			if written != 0x2020:
				self._error()

			self._outfile.close()

			break

		self._has_ended = True

	@_exception_try_shield('_error')
	def _controlled_write(self, x: typing.SupportsBytes) -> None:
		if self._errorstate:
			return

		x = bytes(x)

		if self._limit_write >= 0:
			self._limit_write -= len(x)
			if self._limit_write < 0:
				self._error()
				return

		l = self._outfile.write(x)

		if l != len(x):
			self._error()

	@_exception_try_shield('_error', False)
	def _ensure_seek(self, offset: int) -> bool:
		if self._errorstate:
			return False

		self._outfile.seek(offset)

		if self._outfile.tell() != offset:
			self._error()
			return False

		return True

	@_exception_try_shield('_error')
	def _tmd_end(self, cdn: CDN) -> None:
		off = self._outfile.tell()
		for i, j in cdn.certs.items():
			self._controlled_write(j)

		self._cert_chain_size = self._outfile.tell() - off

		self._block_align_up()

		tik = cdn.ticket
		if tik is not None:
			self._tik_size = len(tik)
			self._controlled_write(tik.installable_export)

			self._block_align_up()

		tmd = cdn.tmd
		if tmd is None:
			self._error()
			return

		self._tmd_size = len(tmd)
		self._controlled_write(tmd)

		self._block_align_up()

		self._content_off = self._outfile.tell()

	@_exception_try_shield('_error')
	def _content_start(self, chunk: ContentChunkRecord) -> typing.Optional[WriterFuncType]:
		if self._curr_content_off >= 0:
			self._error()
			return

		self._limit_write = chunk.size

		self._curr_content_off = self._outfile.tell()

		self._curr_optional = bool(chunk.type & ContentTypeFlags.Optional)

		return self._controlled_write

	@_exception_try_shield('_error')
	def _content_end(self, chunk: ContentChunkRecord) -> None:
		if self._curr_content_off < 0:
			return

		if self._limit_write > 0:
			self._error()
			return

		size = self._outfile.tell() - self._curr_content_off

		if size != chunk.size:
			self._error()
			return

		self._index_field |= 1 << (0xFFFF - chunk.index)

		self._curr_content_off = -1
		self._limit_write = -1
		self._curr_optional = False

		self._block_align_up()

	@_exception_try_shield('_error')
	def notify(
		self,
		stage: CDNDownloadStage,
		chunk: ContentChunkRecord,
		starting: bool,
		http_ret: int,
		cdn: CDN
	) -> typing.Optional[WriterFuncType]:
		if self._end_flag or self._errorstate:
			return None

		if self._last_stage > stage:
			self._error()
			return None

		if self._rollback_opt_content and \
			http_ret == 404 and \
			stage == CDNDownloadStage.CONTENT:
			return None

		if http_ret != 200 and http_ret != 0:
			self._error()
			return None

		self._last_stage = stage

		if stage == CDNDownloadStage.TMD and not starting:
			return self._tmd_end(cdn)

		if stage == CDNDownloadStage.CONTENT and starting:
			return self._content_start(chunk)

		if stage == CDNDownloadStage.CONTENT and not starting:
			return self._content_end(chunk)

		if stage == CDNDownloadStage.END:
			self._end_flag = True
			self._end()

	@_exception_try_shield('_error', True)
	def in_error(self) -> bool:
		return self._errorstate

	def abort(self) -> None:
		self._error()

AssistedCDNDownload.register(CiaCDNBuilder)

class CDNFolderDownload:
	def __init__(self, output: str):
		self._path = pathlib.Path(output).resolve()
		self._path.mkdir(parents=True, exist_ok=True)
		self._errorstate = False
		self._end_flag = False
		self._has_ended = False

		self._limit_write = -1

		self._last_stage = -1

		self._curr_file = None
		self._curr_path = None
		self._is_open = False

	def __del__(self):
		self._end()

	def _error(self) -> None:
		if self._errorstate:
			return

		self._errorstate = True
		if self._is_open:
			try:
				self._curr_file.close()
				self._is_open = False
			except Exception:
				pass
			try:
				self._curr_path.unlink()
			except Exception:
				pass

	@_exception_try_shield('_error')
	def _end(self):
		if self._has_ended:
			return

		if self._is_open:
			self._close_current()

		self._has_ended = True

	@_exception_try_shield('_error')
	def _controlled_write(self, x: typing.SupportsBytes) -> None:
		if self._errorstate:
			return

		x = bytes(x)

		if self._limit_write >= 0:
			self._limit_write -= len(x)
			if self._limit_write < 0:
				self._error()
				return

		l = self._curr_file.write(x)

		if l != len(x):
			self._error()

	@_exception_try_shield('_error')
	def _close_current(self) -> None:
		if self._limit_write > 0:
			self._error()
			return

		if self._is_open:
			self._curr_file.close()

		self._is_open = False

	@_exception_try_shield('_error')
	def _open_current(self, name: str, limit: int = -1) -> typing.Optional[WriterFuncType]:
		if self._is_open:
			self._error()
			return

		self._curr_path = (self._path / name).resolve()
		self._curr_file = self._curr_path.open("wb")
		self._limit_write = limit
		self._is_open = True

		return self._controlled_write

	@_exception_try_shield('_error')
	def notify(
		self,
		stage: CDNDownloadStage,
		chunk: ContentChunkRecord,
		starting: bool,
		http_ret: int,
		cdn: CDN
	) -> typing.Optional[WriterFuncType]:
		if self._end_flag or self._errorstate:
			return None

		if (http_ret != 404 and http_ret != 200 and http_ret != 0) or \
			self._last_stage > stage:
			self._error()
			return

		self._last_stage = stage

		if not starting and http_ret == 404:
			self._close_current()
			self._curr_path.unlink()

		if not starting and stage != CDNDownloadStage.TMD:
			return self._close_current()

		if stage == CDNDownloadStage.CETK:
			return self._open_current("cetk")

		if stage == CDNDownloadStage.TMD and starting:
			return self._open_current("tmd")

		if stage == CDNDownloadStage.TMD and not starting:
			self._close_current()
			w = self._open_current(f"tmd.{cdn.tmd.title_version}")
			if not w:
				self._error()
				return

			w(cdn.tmd_raw)
			return self._close_current()

		if stage == CDNDownloadStage.CONTENT:
			return self._open_current(f"{chunk.id:08X}")

		if stage == CDNDownloadStage.END:
			self._end_flag = True
			self._end()

	@_exception_try_shield('_error', True)
	def in_error(self) -> bool:
		return self._errorstate

	def abort(self) -> None:
		self._error()

AssistedCDNDownload.register(CDNFolderDownload)

class CDNMultiHelper:
	def __init__(
		self,
		helpers: typing.Iterable[AssistedCDNDownload]
	):
		self._helpers = []
		for i in helpers:
			if not isinstance(i, AssistedCDNDownload):
				raise ClassInitError("Non helper in list")
			self._helpers.append(i)

		self._errorstate = False

	def _error(self):
		self._errorstate = True

	@_exception_try_shield('_error')
	def notify(
		self,
		stage: CDNDownloadStage,
		chunk: ContentChunkRecord,
		starting: bool,
		http_ret: int,
		cdn: CDN
	) -> typing.Optional[WriterFuncType]:
		if self._errorstate:
			return

		writers = []

		for i in self._helpers:
			try:
				w = i.notify(stage, chunk, starting, http_ret, cdn)
				if w is not None and callable(w):
					writers.append(w)
			except Exception as e:
				pass

		return MultiWriter(writers).write if writers else None

	@_exception_try_shield('_error', True)
	def in_error(self) -> bool:
		if self._errorstate:
			return True

		for i in self._helpers:
			if not i.in_error():
				return False

		return True # only if all objects are in error state

	def abort(self) -> None:
		for i in self._helpers:
			try:
				i.abort()
			except Exception:
				pass

		self._errorstate = True

AssistedCDNDownload.register(CDNMultiHelper)



def _main():
	import sys
	from collections import OrderedDict

	from .ctr import _main as ctr_main

	# we just support 3ds, for now:tm:
	_mains = OrderedDict()
	_mains['ctr'] = {'func': ctr_main._main, 'desc': '3DS system'}
	_mains['ktr'] = {'func': ctr_main._main, 'desc': 'New 3DS system (ctr alias)'}

	def _print_help():
		print("usage: cleaninty system [args]")
		print("")
		print("systems:")
		for i, j in _mains.items():
			print(f" {i.ljust(10)} {j['desc']}")

	if len(sys.argv) < 2:
		_print_help()
	else:
		_system = _mains.get(sys.argv[1].lower(), None)

		if not _system:
			print("Argument error: invalid system")
			print("")
			_print_help()
		else:
			_system['func'](sys.argv[2:])



import sys, typing, time, math

if sys.hexversion < 0x03070000:
	def monotonic_ns() -> int:
		t = time.monotonic()
		t_F = math.floor(t)
		return t_F * 1000000000 + int((t - t_F) * 1000000000)
else:
	monotonic_ns = time.monotonic_ns

if sys.hexversion < 0x03080000:
	CTRModel_T = str
	LimitKind_T = str
else:
	CTRModel_T = typing.Literal['CTR', 'SPR', 'FTR', 'KTR', 'RED', 'JAN']
	LimitKind_T = typing.Union[ # limit kind
		# don't know at this point, idk what they were thinking
		typing.Literal['PR', 'TR', 'DR', 'SR', 'LR', 'ET'],
		str # max 4 chars
	]



from struct import pack, unpack
import typing

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import ec

from . import certificate
from . import digitalsignature as digsign
from . import otp
from .exception import ClassInitError

__all__ = [
	"APCert"
]

class APCert(certificate.Certificate):
	def __init__(self, ctcert: otp.CTCert, titleid: typing.SupportsInt):
		if not isinstance(ctcert, otp.CTCert):
			raise ClassInitError("APCert excepts CTCert object")

		titleid = int(titleid)

		if titleid < 0 or (titleid & 0xFFFF400000000000) != 0x0004000000000000:
			raise ClassInitError("Invalid titleid")

		try:
			privkey = ec.generate_private_key(ec.SECT233R1(), default_backend())
			pubkeynumbers = privkey.public_key().public_numbers()
		except Exception as e:
			raise ClassInitError("EC Key generation error") from e

		try:
			data = pack(
				">I60x64x64sI64sI30s30s60x",
				digsign.SignatureType.ECC_SHA256,
				b"SigningStaging",
				digsign.KeyType.ECC,
				(f"AP{titleid:016x}").encode(),
				0,
				pubkeynumbers.x.to_bytes(30, 'big'),
				pubkeynumbers.y.to_bytes(30, 'big')
			)
		except Exception as e:
			raise ClassInitError("Packing error") from e

		super().__init__(data)

		try:
			self.reissue(ctcert)
			self.verify(ctcert)
		except Exception as e:
			raise ClassInitError("Could not sign or verify the generated APCert") from e

		if not self.set_private_key(privkey.private_numbers()):
			raise ClassInitError("APCert private key was not successfully loaded!")

		self._ap_title_id = titleid

	@property
	def ap_title_id(self) -> int:
		return self._ap_title_id



from struct import unpack
import typing, hashlib, os, pathlib, json

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
import cryptography.hazmat.primitives.ciphers as crypto_ciphers

from .environment import find_in_config_dirs
from . import constants

__all__ = [
	"get_b9_x",
	"get_b9_y",
	"get_b9_n",
	"get_rsa_key",
	"scramble_keys",
	"unscramble_key_with_x",
	"unscramble_key_with_y",
	"register_additional_b9_paths"
]

_b9_keydata = {
	'retail': {
		'x': (None,)*64, 'y': (None,)*64, 'n': (None,)*64,
		'otp': {'key': None, 'iv': None},
		'keyblob': None,
		'rsa': (None,)*11,
		'rsa_priv': [None,]*8
	},
	'dev': {
		'x': (None,)*64, 'y': (None,)*64, 'n': (None,)*64,
		'otp': {'key': None, 'iv': None},
		'keyblob': None,
		'rsa': (None,)*11,
		'rsa_priv': [None,]*8
	}
}

_b9_prot_sha512_hash = 'd1f10601193b4154d8c7667102f7c5e270fea49d50b04c6f7e374dbf15937c1a5568236bb551a5cb73bf789c9454c2727118c5eb849f0728561273b684ca1ae7'

_loaded_b9 = False

_additional_b9_paths: typing.List[str] = []

def _get_b9_env() -> typing.List[str]:
	env = os.environ.get('BOOT9_PATH', None)
	if not env:
		return []

	try:
		path = pathlib.Path(env).resolve()
		env = str(path) if path.is_file() else []

	except Exception:
		env = []

	return env

def _load_b9_keys() -> bool:
	global _loaded_b9
	if _loaded_b9:
		return True

	paths = find_in_config_dirs('boot9.bin')
	paths += find_in_config_dirs('boot9_protected.bin')
	paths += _get_b9_env() 
	paths += _additional_b9_paths
	paths = list(set(paths))

	got_b9 = False

	for i in paths:
		try:
			with open(i, 'rb') as file:
				data = file.read(0x10000)

			if len(data) == 0x10000:
				data = data[0x8000:]
			else:
				data = data[:0x8000]
			
			if len(data) < 0x8000:
				continue

			if hashlib.sha512(data).hexdigest() != _b9_prot_sha512_hash:
				continue

			got_b9 = True
			break
			
		except Exception:
			continue

	if not got_b9:
		return False

	be = lambda x: int.from_bytes(x, 'big')

	e = 0x10001

	rsa_raw_retail, rsa_raw_dev, \
	otp_key_retail, otp_iv_retail, otp_key_dev, otp_iv_dev, \
	keyblob_raw_retail, keyblob_raw_dev = unpack('12512x 4864s 4864s 16s 16s 16s 16s 320x 1024s 1024s 8096x', data)

	keydata = {
		'retail': {
			'x': None, 'y': None, 'n': None,
			'otp': {'key': be(otp_key_retail), 'iv': be(otp_iv_retail)},
			'keyblob': keyblob_raw_retail,
			'rsa': None,
			'rsa_priv': [None,]*8
		},
		'dev': {
			'x': None, 'y': None, 'n': None,
			'otp': {'key': be(otp_key_dev), 'iv': be(otp_iv_dev)},
			'keyblob': keyblob_raw_dev,
			'rsa': None,
			'rsa_priv': [None,]*8
		}
	}

	rsa_keys = [[], []]

	for i, k in enumerate((unpack('256s'*19, rsa_raw_retail), unpack('256s'*19, rsa_raw_dev))):
		try:
			for j in range(0,3):
				key = rsa.RSAPublicNumbers(e, be(k[j]))
				key.public_key(default_backend()) # validate key by trying to load

				rsa_keys[i].append(key)

			for j in range(3,19,2):
				n = be(k[j])
				d = be(k[j+1])

				key = (rsa.RSAPublicNumbers(e, n), d)
				key[0].public_key(default_backend()) # validate key by trying to load

				rsa_keys[i].append(key)

		except Exception:
			return False

	keydata['retail']['rsa'] = tuple(rsa_keys[0])
	keydata['dev']['rsa'] = tuple(rsa_keys[1])
	del rsa_keys

	unpacked_keys = [
		[
			[None]*64,
			[None]*64,
			[None]*64
		],
		[
			[None]*64,
			[None]*64,
			[None]*64
		]
	]

	for i, k in enumerate((unpack('16s'*36, keyblob_raw_retail[368:-80]), unpack('16s'*36, keyblob_raw_dev[368:-80]))):
		# x
		unpacked_keys[i][0][0x2C:0x30] = [be(k[0])]*4
		unpacked_keys[i][0][0x30:0x34] = [be(k[1])]*4
		unpacked_keys[i][0][0x34:0x38] = [be(k[2])]*4
		unpacked_keys[i][0][0x38:0x3C] = [be(k[3])]*4
		unpacked_keys[i][0][0x3C:0x40] = [be(k[j]) for j in range(4,8)]

		# y
		unpacked_keys[i][1][0x04:0x0C] = [be(k[j]) for j in range(8,16)]

		# n
		unpacked_keys[i][2][0x0C:0x10] = [be(k[16])]*4
		unpacked_keys[i][2][0x10] = be(k[17])
		unpacked_keys[i][2][0x14:0x18] = [be(k[j]) for j in range(18,22)]
		unpacked_keys[i][2][0x18:0x1C] = [be(k[22])]*4
		unpacked_keys[i][2][0x1C:0x20] = [be(k[23])]*4
		unpacked_keys[i][2][0x20:0x24] = [be(k[24])]*4
		unpacked_keys[i][2][0x24] = unpacked_keys[i][2][0x28] = be(k[25])
		unpacked_keys[i][2][0x29:0x2C] = [be(k[j]) for j in range(26,29)]
		unpacked_keys[i][2][0x2C:0x30] = [be(k[29])]*4
		unpacked_keys[i][2][0x30:0x34] = [be(k[30])]*4
		unpacked_keys[i][2][0x34:0x38] = [be(k[31])]*4
		unpacked_keys[i][2][0x38] = unpacked_keys[i][2][0x3C] = be(k[32])
		unpacked_keys[i][2][0x3D:0x40] = [be(k[j]) for j in range(33,36)]

	keydata['retail']['x'] = tuple(unpacked_keys[0][0])
	keydata['retail']['y'] = tuple(unpacked_keys[0][1])
	keydata['retail']['n'] = tuple(unpacked_keys[0][2])
	keydata['dev']['x'] = tuple(unpacked_keys[1][0])
	keydata['dev']['y'] = tuple(unpacked_keys[1][1])
	keydata['dev']['n'] = tuple(unpacked_keys[1][2])
	del unpacked_keys

	global _b9_keydata
	_b9_keydata = keydata

	_loaded_b9 = True

	return True

def _ensure_b9(fail_return_value: typing.Any):
	def inner(func):
		def wrapper(*args, **kwargs):
			if _load_b9_keys():
				return func(*args, **kwargs)
			return fail_return_value
		return wrapper
	return inner

@_ensure_b9(None)
def _decrypt_otp(otp: typing.SupportsBytes, dev: bool) -> typing.Optional[bytes]:
	otp = bytes(otp)[:0x100]
	keydata = _b9_keydata['dev' if dev else 'retail']['otp']
	decryptor = crypto_ciphers.Cipher(
		crypto_ciphers.algorithms.AES(keydata['key'].to_bytes(16, 'big')),
		crypto_ciphers.modes.CBC(keydata['iv'].to_bytes(16, 'big')),
		default_backend()
	).decryptor()
	dec_otp = decryptor.update(otp) + decryptor.finalize()
	return dec_otp

@_ensure_b9(None)
def _encrypt_otp(otp: typing.SupportsBytes, dev: bool) -> typing.Optional[bytes]:
	otp = bytes(otp)[:0x100]
	keydata = _b9_keydata['dev' if dev else 'retail']['otp']
	encryptor = crypto_ciphers.Cipher(
		crypto_ciphers.algorithms.AES(keydata['key'].to_bytes(16, 'big')),
		crypto_ciphers.modes.CBC(keydata['iv'].to_bytes(16, 'big')),
		default_backend()
	).encryptor()
	enc_otp = encryptor.update(otp) + encryptor.finalize()
	return enc_otp

@_ensure_b9(None)
def _gen_console_keys(otp_blob: typing.SupportsBytes, dev: bool) -> typing.Optional[typing.Iterable[typing.Optional[int]]]:
	keyblob = _b9_keydata['dev' if dev else 'retail']['keyblob']
	be = lambda x: int.from_bytes(x, 'big')

	hashing_blob = bytes(otp_blob)[:0x1C] + keyblob[:36]
	blob_hash = hashlib.sha256(hashing_blob).digest()

	key = scramble_keys(be(blob_hash[0:16]), be(blob_hash[16:32]))
	if key is None:
		return None

	key = key.to_bytes(16, 'big')

	genblobs = (
		unpack('36x16s64s', keyblob[0:116]),
		unpack('36x16s64s', keyblob[116:232]),
		unpack('36x16s64s', keyblob[184:300]),
		unpack('36x16s64s', keyblob[300:416])
	)

	make_encryptor = lambda iv: crypto_ciphers.Cipher(
		crypto_ciphers.algorithms.AES(key),
		crypto_ciphers.modes.CBC(iv),
		default_backend()
	).encryptor()
	encrypt = lambda encryptor, data: encryptor.update(data) + encryptor.finalize()

	key_blobs = (
		encrypt(make_encryptor(genblobs[0][0]), genblobs[0][1]),
		encrypt(make_encryptor(genblobs[1][0]), genblobs[1][1]),
		encrypt(make_encryptor(genblobs[2][0]), genblobs[2][1]),
		encrypt(make_encryptor(genblobs[3][0]), genblobs[3][1])
	)

	keys = [None]*64

	keys[0x04:0x08] = [be(key_blobs[0][0:16])]*4
	keys[0x08:0x0C] = [be(key_blobs[0][16:32])]*4
	keys[0x0C:0x10] = [be(key_blobs[0][32:48])]*4
	keys[0x10] = be(key_blobs[0][48:64])
	keys[0x14:0x18] = [be(key_blobs[1][i:i+16]) for i in range(0,64,16)]
	keys[0x18:0x1C] = [be(key_blobs[2][0:16])]*4
	keys[0x1C:0x20] = [be(key_blobs[2][16:32])]*4
	keys[0x20:0x24] = [be(key_blobs[2][32:48])]*4
	keys[0x24] = be(key_blobs[2][48:64])
	keys[0x28:0x2C] = [be(key_blobs[3][i:i+16]) for i in range(0,64,16)]

	return tuple(keys)

@_ensure_b9(None)
def get_b9_x(slot: int, dev: bool) -> typing.Optional[int]:
	if not 0 <= slot < 64:
		return None
	return _b9_keydata['dev' if dev else 'retail']['x'][slot]

@_ensure_b9(None)
def get_b9_y(slot: int, dev: bool) -> typing.Optional[int]:
	if not 0 <= slot < 64:
		return None
	return _b9_keydata['dev' if dev else 'retail']['y'][slot]

@_ensure_b9(None)
def get_b9_n(slot: int, dev: bool) -> typing.Optional[int]:
	if not 0 <= slot < 64:
		return None
	return _b9_keydata['dev' if dev else 'retail']['n'][slot]

@_ensure_b9(None)
def get_rsa_key(index: int, dev: bool) -> typing.Union[
	rsa.RSAPublicKey, rsa.RSAPublicKeyWithSerialization,
	rsa.RSAPrivateKey, rsa.RSAPrivateKeyWithSerialization,
	None
]:
	global _b9_keydata

	if not 0 <= index < 11:
		return None

	sel = 'dev' if dev else 'retail'

	r = _b9_keydata[sel]['rsa'][index]

	if index < 3:
		return rsa.RSAPublicNumbers(
			r.e, r.n
		).public_key(default_backend())

	pub = r[0]
	p_index = index-3
	priv = _b9_keydata[sel]['rsa_priv'][p_index]

	if priv is None:
		n, e, d = (pub.n, pub.e, r[1])

		p, q = rsa.rsa_recover_prime_factors(n, e, d)
		p, q = (p, q) if p > q else (q, p)
		dmp1 = rsa.rsa_crt_dmp1(d, p)
		dmq1 = rsa.rsa_crt_dmq1(d, q)
		iqmp = rsa.rsa_crt_iqmp(p, q)
		key = rsa.RSAPrivateNumbers(
			p, q, d, dmp1, dmq1, iqmp,
			rsa.RSAPublicNumbers(e, n)
		)
		ret = key.private_key(default_backend())

		_b9_keydata[sel]['rsa_priv'][p_index] = key

		return ret
	else:
		r = priv
		pub = r.public_numbers

	return rsa.RSAPrivateNumbers(
		r.p, r.q, r.d, r.dmp1, r.dmq1, r.iqmp,
		rsa.RSAPublicNumbers(pub.e, pub.n)
	).private_key(default_backend())

constants._load_const_c()

def scramble_keys(x: int, y: int) -> typing.Optional[int]:
	if constants._aes_const_c is None:
		constants._load_const_c()
		if constants._aes_const_c is None:
			return None

	mask = (1<<128)-1
	x &= mask
	y &= mask
	tmp = (x << 2) | (x >> 126)
	tmp ^= y
	tmp += constants._aes_const_c
	tmp &= mask
	tmp = (tmp << 87) | (tmp >> 41)
	return tmp & mask

def unscramble_key_with_x(n: int, x: int) -> typing.Optional[int]:
	if constants._aes_const_c is None:
		constants._load_const_c()
		if constants._aes_const_c is None:
			return None

	mask = (1<<128)-1
	n &= mask
	x &= mask
	tmp = (n << 41) | (n >> 87)
	tmp -= constants._aes_const_c
	tmp &= mask
	tmp ^= (x << 2) | (x >> 126)
	return tmp & mask

def unscramble_key_with_y(n: int, y: int) -> typing.Optional[int]:
	if constants._aes_const_c is None:
		constants._load_const_c()
		if constants._aes_const_c is None:
			return None

	mask = (1<<128)-1
	n &= mask
	y &= mask
	tmp = (n << 41) | (n >> 87)
	tmp -= constants._aes_const_c
	tmp &= mask
	tmp ^= y
	tmp = (tmp << 126) | (tmp >> 2)
	return tmp & mask

def register_additional_b9_paths(*args):
	global _additional_b9_paths

	paths = []

	for i in args:
		try:
			path = pathlib.Path(i).resolve()
			if not path.is_file():
				continue

			paths.append(str(path))

		except Exception:
			pass

	_additional_b9_paths = list(set(_additional_b9_paths + paths))
